# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import glob
import hashlib
import json
import logging
import os
import sys
from functools import wraps
from unittest.mock import patch

import numpy as np
import pytest
from anemoi.utils.testing import get_test_archive
from anemoi.utils.testing import get_test_data
from anemoi.utils.testing import skip_if_offline
from earthkit.data import from_source as original_from_source

from anemoi.datasets import open_dataset
from anemoi.datasets.create.testing import create_dataset
from anemoi.datasets.data.stores import open_zarr

HERE = os.path.dirname(__file__)
# find_yamls
NAMES = sorted([os.path.basename(path).split(".")[0] for path in glob.glob(os.path.join(HERE, "*.yaml"))])
SKIP = ["recentre"]
SKIP += ["accumulation"]  # test not in s3 yet
SKIP += ["regrid"]
NAMES = [name for name in NAMES if name not in SKIP]
assert NAMES, "No yaml files found in " + HERE


def mockup_from_source(func: callable) -> callable:
    """Decorator to mock the `from_source` function from the `earthkit.data` module.

    Parameters
    ----------
    func : function
        The function to be wrapped.

    Returns
    -------
    function
        The wrapped function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        with patch("earthkit.data.from_source", _from_source):
            return func(*args, **kwargs)

    return wrapper


class LoadSource:
    """Class to load data sources and handle mockup data."""

    def filename(self, args: tuple, kwargs: dict) -> str:
        """Generate a filename based on the arguments and keyword arguments.

        Parameters
        ----------
        args : tuple
            The positional arguments.
        kwargs : dict
            The keyword arguments.

        Returns
        -------
        str
            The generated filename.
        """
        string = json.dumps([args, kwargs], sort_keys=True, default=str)
        h = hashlib.md5(string.encode("utf8")).hexdigest()
        return h + ".grib"

    def get_data(self, args: tuple, kwargs: dict, path: str) -> None:
        """Retrieve data and save it to the specified path.

        Parameters
        ----------
        args : tuple
            The positional arguments.
        kwargs : dict
            The keyword arguments.
        path : str
            The path to save the data.

        Raises
        ------
        ValueError
            If the test data is missing.
        """
        upload_path = os.path.realpath(path + ".to_upload")
        ds = original_from_source("mars", *args, **kwargs)
        ds.save(upload_path)
        print(f"Mockup: Saving to {upload_path} for {args}, {kwargs}")
        print()
        print("⚠️ To upload the test data, run this:")
        path = os.path.relpath(upload_path, os.getcwd())
        name = os.path.basename(upload_path).replace(".to_upload", "")
        print(f"scp {path} data@anemoi.ecmwf.int:public/anemoi-datasets/create/mock-mars/{name}")
        print()
        exit(1)
        raise ValueError("Test data is missing")

    def mars(self, args: tuple, kwargs: dict) -> object:
        """Load data from the MARS archive.

        Parameters
        ----------
        args : tuple
            The positional arguments.
        kwargs : dict
            The keyword arguments.

        Returns
        -------
        object
            The loaded data source.
        """

        name = self.filename(args, kwargs)

        try:
            return original_from_source("file", get_test_data(f"anemoi-datasets/create/mock-mars/{name}"))
        except RuntimeError:
            raise  # If offline
        except Exception:
            self.get_data(args, kwargs, name)

    def __call__(self, name: str, *args: tuple, **kwargs: dict) -> object:
        """Call the appropriate method based on the data source name.

        Parameters
        ----------
        name : str
            The name of the data source.
        args : tuple
            The positional arguments.
        kwargs : dict
            The keyword arguments.

        Returns
        -------
        object
            The loaded data source.
        """
        if name == "mars":
            return self.mars(args, kwargs)

        return original_from_source(name, *args, **kwargs)


_from_source = LoadSource()


def compare_dot_zattrs(a: dict, b: dict, path: str, errors: list) -> None:
    """Compare the attributes of two Zarr datasets.

    Parameters
    ----------
    a : dict
        The attributes of the first dataset.
    b : dict
        The attributes of the second dataset.
    path : str
        The current path in the attribute hierarchy.
    errors : list
        The list to store error messages.
    """
    if isinstance(a, dict):
        a_keys = list(a.keys())
        b_keys = list(b.keys())
        for k in set(a_keys) | set(b_keys):
            if k not in a_keys:
                errors.append(f"❌ {path}.{k} : missing key (only in reference)")
                continue
            if k not in b_keys:
                errors.append(f"❌ {path}.{k} : additional key (missing in reference)")
                continue
            if k in [
                "timestamp",
                "uuid",
                "latest_write_timestamp",
                "history",
                "provenance",
                "provenance_load",
                "description",
                "config_path",
                "total_size",
            ]:
                if type(a[k]) is not type(b[k]):
                    errors.append(f"❌ {path}.{k} : type differs {type(a[k])} != {type(b[k])}")
                continue

            compare_dot_zattrs(a[k], b[k], f"{path}.{k}", errors)

        return

    if isinstance(a, list):
        if len(a) != len(b):
            errors.append(f"❌ {path} : lengths are different {len(a)} != {len(b)}")
            return

        for i, (v, w) in enumerate(zip(a, b)):
            compare_dot_zattrs(v, w, f"{path}.{i}", errors)

        return

    if type(a) is not type(b):
        msg = f"❌ {path} actual != expected : {a} ({type(a)}) != {b} ({type(b)})"
        errors.append(msg)
        return

    if a != b:
        msg = f"❌ {path} actual != expected : {a} != {b}"
        errors.append(msg)


def compare_datasets(a: object, b: object) -> None:
    """Compare two datasets.

    Parameters
    ----------
    a : object
        The first dataset.
    b : object
        The second dataset.

    Raises
    ------
    AssertionError
        If the datasets do not match.
    """
    assert a.shape == b.shape, (a.shape, b.shape)
    assert (a.dates == b.dates).all(), (a.dates, b.dates)
    for a_, b_ in zip(a.variables, b.variables):
        assert a_ == b_, (a, b)
    assert a.missing == b.missing, "Missing are different"

    for i_date, date in zip(range(a.shape[0]), a.dates):
        if i_date in a.missing:
            continue
        for i_param in range(a.shape[1]):
            param = a.variables[i_param]
            assert param == b.variables[i_param], (
                date,
                param,
                a.variables[i_param],
                b.variables[i_param],
            )
            a_ = a[i_date, i_param]
            b_ = b[i_date, i_param]
            assert a.shape == b.shape, (date, param, a.shape, b.shape)

            a_nans = np.isnan(a_)
            b_nans = np.isnan(b_)
            assert np.all(a_nans == b_nans), (date, param, "nans are different")

            a_ = np.where(a_nans, 0, a_)
            b_ = np.where(b_nans, 0, b_)

            delta = a_ - b_
            max_delta = np.max(np.abs(delta))
            abs_error = np.abs(a_ - b_)
            rel_error = np.abs(a_ - b_) / (np.abs(b_) + 1e-10)  # Avoid division by zero
            assert max_delta == 0.0, (date, param, a_, b_, a_ - b_, max_delta, np.max(abs_error), np.max(rel_error))


def compare_statistics(ds1: object, ds2: object) -> None:
    """Compare the statistics of two datasets.

    Parameters
    ----------
    ds1 : object
        The first dataset.
    ds2 : object
        The second dataset.

    Raises
    ------
    AssertionError
        If the statistics do not match.
    """
    vars1 = ds1.variables
    vars2 = ds2.variables
    assert len(vars1) == len(vars2)
    for v1, v2 in zip(vars1, vars2):
        idx1 = ds1.name_to_index[v1]
        idx2 = ds2.name_to_index[v2]
        assert (ds1.statistics["mean"][idx1] == ds2.statistics["mean"][idx2]).all()
        assert (ds1.statistics["stdev"][idx1] == ds2.statistics["stdev"][idx2]).all()
        assert (ds1.statistics["maximum"][idx1] == ds2.statistics["maximum"][idx2]).all()
        assert (ds1.statistics["minimum"][idx1] == ds2.statistics["minimum"][idx2]).all()


class Comparer:
    """Class to compare datasets and their metadata.

    Parameters
    ----------
    name : str
        The name of the dataset.
    output_path : str, optional
        The path to the output dataset.
    reference_path : str, optional
        The path to the reference dataset.
    """

    def __init__(self, name: str, output_path: str = None, reference_path: str = None) -> None:
        """Initialize the Comparer instance.

        Parameters
        ----------
        name : str
            The name of the dataset.
        output_path : str, optional
            The path to the output dataset.
        reference_path : str, optional
            The path to the reference dataset.
        """
        self.name = name
        self.output_path = output_path or os.path.join(name + ".zarr")
        self.reference_path = reference_path
        print(f"Comparing {self.output_path} and {self.reference_path}")

        self.z_output = open_zarr(self.output_path)
        self.z_reference = open_zarr(self.reference_path)

        self.z_reference["data"]
        self.ds_output = open_dataset(self.output_path)
        self.ds_reference = open_dataset(self.reference_path)

    def compare(self) -> None:
        """Compare the output dataset with the reference dataset.

        Raises
        ------
        AssertionError
            If the datasets or their metadata do not match.
        """
        errors = []
        compare_dot_zattrs(dict(self.z_output.attrs), dict(self.z_reference.attrs), "metadata", errors)
        if errors:
            print("Comparison failed")
            print("\n".join(errors))

        if errors:
            print()

            print()
            print("⚠️ To update the reference data, run this:")
            print("cd " + os.path.dirname(self.output_path))
            base = os.path.basename(self.output_path)
            print(f"tar zcf {base}.tgz {base}")
            print(f"scp {base}.tgz data@anemoi.ecmwf.int:public/anemoi-datasets/create/mock-mars/")
            print()
            raise AssertionError("Comparison failed")

        compare_datasets(self.ds_output, self.ds_reference)
        compare_statistics(self.ds_output, self.ds_reference)
        # do not compare tendencies statistics yet, as we don't know yet if they should stay


@skip_if_offline
@pytest.mark.parametrize("name", NAMES)
@mockup_from_source
def test_run(name: str) -> None:
    """Run the test for the specified dataset.

    Parameters
    ----------
    name : str
        The name of the dataset.

    Raises
    ------
    AssertionError
        If the comparison fails.
    """
    config = os.path.join(HERE, name + ".yaml")
    output = os.path.join(HERE, name + ".zarr")
    is_test = False

    create_dataset(config=config, output=output, delta=["12h"], is_test=is_test)

    directory = get_test_archive(f"anemoi-datasets/create/mock-mars/{name}.zarr.tgz")
    reference = os.path.join(directory, name + ".zarr")

    Comparer(name, output_path=output, reference_path=reference).compare()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    if len(sys.argv) > 1:
        names = sys.argv[1:]
    else:
        names = NAMES

    for name in names:
        logging.info(f"Running test for {name}")
        try:
            test_run(name)
        except AssertionError:
            pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property
from typing import List

from earthkit.data import FieldList

from .misc import assert_fieldlist
from .result import Result
from .trace import trace_datasource

LOG = logging.getLogger(__name__)


class EmptyResult(Result):
    """Class to represent an empty result in the dataset creation process."""

    empty = True

    def __init__(self, context: object, action_path: list, dates: object) -> None:
        """Initializes an EmptyResult instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : list
            The action path.
        dates : object
            The dates object.
        """
        super().__init__(context, action_path + ["empty"], dates)

    @cached_property
    @assert_fieldlist
    @trace_datasource
    def datasource(self) -> FieldList:
        """Returns an empty datasource."""
        from earthkit.data import from_source

        return from_source("empty")

    @property
    def variables(self) -> List[str]:
        """Returns an empty list of variables."""
        return []



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import time
from typing import Any

from anemoi.utils.humanize import seconds_to_human

from anemoi.datasets.commands.create import task

from . import Command

LOG = logging.getLogger(__name__)


class Load(Command):
    """Create a dataset, step by step."""

    internal = True
    timestamp = True

    def add_arguments(self, subparser: Any) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        subparser : Any
            The command parser.
        """
        subparser.add_argument("--parts", nargs="+", help="Only load the specified parts of the dataset.")
        # subparser.add_argument(
        #        "--delta",
        #        help="Compute statistics tendencies on a given time delta, if possible. Must be a multiple of the frequency.",
        #    )

        subparser.add_argument("path", help="Path to store the created data.")
        subparser.add_argument("--cache", help="Location to store the downloaded data.", metavar="DIR")
        subparser.add_argument("--trace", action="store_true")

    def run(self, args: Any) -> None:
        """Run the command.

        Parameters
        ----------
        args : Any
            The command arguments.
        """
        options = vars(args)
        options.pop("command")
        now = time.time()
        step = "load"

        if "version" in options:
            options.pop("version")

        if "debug" in options:
            options.pop("debug")
            task(step, options)

        LOG.info(f"Create step '{step}' completed in {seconds_to_human(time.time()-now)}")


command = Load



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import logging
import os
from typing import Union

import zarr

LOG = logging.getLogger(__name__)


def fix_order_by(order_by: Union[dict, list]) -> list[dict]:
    """Fix the order_by attribute to ensure it is a list of dictionaries.

    Parameters
    ----------
    order_by : dict or list
        The order_by attribute to fix.

    Returns
    -------
    list[dict]
        The fixed order_by attribute.
    """
    if isinstance(order_by, list):
        return order_by

    assert isinstance(order_by, dict), order_by
    assert len(order_by) <= 3, order_by
    lst = []
    lst.append({"valid_datetime": order_by["valid_datetime"]})
    lst.append({"param_level": order_by["param_level"]})
    lst.append({"number": order_by["number"]})
    return lst


def fix_history(history: list[dict]) -> list[dict]:
    """Fix the history attribute by removing specific actions.

    Parameters
    ----------
    history : list[dict]
        The history attribute to fix.

    Returns
    -------
    list[dict]
        The fixed history attribute.
    """
    new = history
    new = [d for d in new if d.get("action") != "loading_data_start"]
    new = [d for d in new if d.get("action") != "loading_data_end"]
    return new


def fix_provenance(provenance: dict) -> dict:
    """Fix the provenance attribute by adding missing fields and removing unnecessary ones.

    Parameters
    ----------
    provenance : dict
        The provenance attribute to fix.

    Returns
    -------
    dict
        The fixed provenance attribute.
    """
    if "python" not in provenance:
        provenance["python"] = provenance["platform"]["python_version"]

    for q in (
        "args",
        "config_paths",
        "executable",
        "gpus",
        "platform",
        "python_path",
        "assets",
    ):
        if q in provenance:
            del provenance[q]

    for k, v in list(provenance["module_versions"].items()):
        if v.startswith("<"):
            del provenance["module_versions"][k]
        if v.startswith("/"):
            provenance["module_versions"][k] = os.path.join("...", os.path.basename(v))

    for k, v in list(provenance["git_versions"].items()):
        LOG.debug(k, v)
        modified_files = v["git"].get("modified_files", [])
        untracked_files = v["git"].get("untracked_files", [])
        if not isinstance(modified_files, int):
            modified_files = len(modified_files)
        if not isinstance(untracked_files, int):
            untracked_files = len(untracked_files)
        provenance["git_versions"][k] = dict(
            git={
                "sha1": v["git"]["sha1"],
                "modified_files": modified_files,
                "untracked_files": untracked_files,
            }
        )

    LOG.debug(json.dumps(provenance, indent=2))
    # assert False
    return provenance


def apply_patch(path: str, verbose: bool = True, dry_run: bool = False) -> None:
    """Apply a patch to the dataset at the given path.

    Parameters
    ----------
    path : str
        The path to the dataset.
    verbose : bool, optional
        Whether to log detailed information. Defaults to True.
    dry_run : bool, optional
        If True, do not actually apply the patch. Defaults to False.
    """
    LOG.debug("====================")
    LOG.debug(f"Patching {path}")
    LOG.debug("====================")

    try:
        attrs = zarr.open(path, mode="r").attrs.asdict()
    except zarr.errors.PathNotFoundError as e:
        LOG.error(f"Failed to open {path}")
        LOG.error(e)
        exit(0)

    FIXES = {
        "history": fix_history,
        "provenance_load": fix_provenance,
        "provenance_statistics": fix_provenance,
        "order_by": fix_order_by,
    }
    REMOVE = ["_create_yaml_config"]

    before = json.dumps(attrs, sort_keys=True)

    fixed_attrs = {}
    for k, v in attrs.items():
        v = attrs[k]
        if k in REMOVE:
            LOG.info(f"✅ Remove {k}")
            continue

        if k not in FIXES:
            assert not k.startswith("provenance"), f"[{k}]"
            LOG.debug(f"✅ Don't fix {k}")
            fixed_attrs[k] = v
            continue

        new_v = FIXES[k](v)
        if json.dumps(new_v, sort_keys=True) != json.dumps(v, sort_keys=True):
            LOG.info(f"✅ Fix {k}")
            if verbose:
                LOG.info(f"  Before : {k}= {v}")
                LOG.info(f"  After  : {k}= {new_v}")
        else:
            LOG.debug(f"✅ Unchanged {k}")
        fixed_attrs[k] = new_v

    if dry_run:
        return
    z = zarr.open(path, mode="r+")

    for k in list(z.attrs.keys()):
        if k not in fixed_attrs:
            del z.attrs[k]
    for k, v in fixed_attrs.items():
        z.attrs[k] = v

    after = json.dumps(z.attrs.asdict(), sort_keys=True)
    if before != after:
        LOG.info("Dataset changed by patch")

    assert json.dumps(z.attrs.asdict(), sort_keys=True) == json.dumps(fixed_attrs, sort_keys=True)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import Optional

from anemoi.utils.dates import as_datetime
from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.utils.projections import Projection

LOG = logging.getLogger(__name__)


class _MDMapping:
    """A class to handle metadata mapping for variables.

    Attributes
    ----------
    variable : Any
        The variable to map.
    time : Any
        The time associated with the variable.
    mapping : Dict[str, str]
        A dictionary mapping keys to variable names.
    """

    def __init__(self, variable: Any) -> None:
        """Initialize the _MDMapping class.

        Parameters
        ----------
        variable : Any
            The variable to map.
        """
        self.variable = variable
        self.time = variable.time
        self.mapping = dict()
        # Aliases

    def _from_user(self, key: str) -> str:
        """Get the internal key corresponding to a user-provided key.

        Parameters
        ----------
        key : str
            The user-provided key.

        Returns
        -------
        str
            The internal key corresponding to the user-provided key.
        """
        return self.mapping.get(key, key)

    def from_user(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Convert user-provided keys to internal keys.

        Parameters
        ----------
        kwargs : Dict[str, Any]
            A dictionary of user-provided keys and values.

        Returns
        -------
        Dict[str, Any]
            A dictionary with internal keys and original values.
        """
        return {self._from_user(k): v for k, v in kwargs.items()}

    def __repr__(self) -> str:
        """Return a string representation of the _MDMapping object.

        Returns
        -------
        str
            String representation of the _MDMapping object.
        """
        return f"MDMapping({self.mapping})"

    def fill_time_metadata(self, field: Any, md: Dict[str, Any]) -> None:
        """Fill the time metadata for a field.

        Parameters
        ----------
        field : Any
            The field to fill metadata for.
        md : Dict[str, Any]
            The metadata dictionary to update.
        """
        valid_datetime = self.variable.time.fill_time_metadata(field._md, md)
        if valid_datetime is not None:
            md["valid_datetime"] = as_datetime(valid_datetime).isoformat()


class XArrayMetadata(RawMetadata):
    """A class to handle metadata for XArray fields.

    Attributes
    ----------
    LS_KEYS : List[str]
        List of keys for the metadata.
    NAMESPACES : List[str]
        List of namespaces for the metadata.
    MARS_KEYS : List[str]
        List of MARS keys for the metadata.
    """

    LS_KEYS = ["variable", "level", "valid_datetime", "units"]
    NAMESPACES = ["default", "mars"]
    MARS_KEYS = ["param", "step", "levelist", "levtype", "number", "date", "time"]

    def __init__(self, field: Any) -> None:
        """Initialize the XArrayMetadata class.

        Parameters
        ----------
        field : Any
            The field to extract metadata from.
        """
        self._field = field
        md = field._md.copy()
        self._mapping = _MDMapping(field.owner)
        self._mapping.fill_time_metadata(field, md)
        super().__init__(md)

    @cached_property
    def geography(self) -> "XArrayFieldGeography":
        """Get the geography information for the field."""
        return XArrayFieldGeography(self._field, self._field.owner.grid)

    def as_namespace(self, namespace: Optional[str] = None) -> Dict[str, Any]:
        """Get the metadata as a specific namespace.

        Parameters
        ----------
        namespace : Optional[str]
            The namespace to use.

        Returns
        -------
        Dict[str, Any]
            The metadata in the specified namespace.
        """
        if not isinstance(namespace, str) and namespace is not None:
            raise TypeError("namespace must be a str or None")

        if namespace == "default" or namespace == "" or namespace is None:
            return dict(self)

        elif namespace == "mars":
            return self._as_mars()

    def _as_mars(self) -> Dict[str, Any]:
        """Get the metadata as MARS namespace.

        Returns
        -------
        Dict[str, Any]
            The metadata in the MARS namespace.
        """
        return {}

    def _base_datetime(self) -> Optional[datetime.datetime]:
        """Get the base datetime for the field.

        Returns
        -------
        Optional[datetime.datetime]
            The base datetime for the field.
        """
        return self._field.forecast_reference_time

    def _valid_datetime(self) -> Optional[datetime.datetime]:
        """Get the valid datetime for the field.

        Returns
        -------
        Optional[datetime.datetime]
            The valid datetime for the field.
        """
        return self._get("valid_datetime")

    def get(self, key: str, astype: Optional[type] = None, **kwargs: Any) -> Any:
        """Get a metadata value by key.

        Parameters
        ----------
        key : str
            The key to get the value for.
        astype : Optional[type]
            The type to cast the value to.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Any
            The value for the specified key, optionally cast to the specified type.
        """
        if key in self._d:
            if astype is not None:
                return astype(self._d[key])
            return self._d[key]

        key = self._mapping._from_user(key)

        return super().get(key, astype=astype, **kwargs)


class XArrayFieldGeography(Geography):
    """A class to handle geography information for XArray fields.

    Attributes
    ----------
    _field : Any
        The field to extract geography information from.
    _grid : Any
        The grid associated with the field.
    """

    def __init__(self, field: Any, grid: Any) -> None:
        """Initialize the XArrayFieldGeography class.

        Parameters
        ----------
        field : Any
            The field to extract geography information from.
        grid : Any
            The grid associated with the field.
        """
        self._field = field
        self._grid = grid

    def _unique_grid_id(self) -> None:
        """Get the unique grid ID.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError()

    def bounding_box(self) -> None:
        """Get the bounding box for the field.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError()
        # return BoundingBox(north=self.north, south=self.south, east=self.east, west=self.west)

    def gridspec(self) -> None:
        """Get the grid specification for the field.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError()

    def latitudes(self, dtype: Optional[type] = None) -> Any:
        """Get the latitudes for the field.

        Parameters
        ----------
        dtype : Optional[type]
            The type to cast the latitudes to.

        Returns
        -------
        Any
            The latitudes for the field.
        """
        result = self._grid.latitudes
        if dtype is not None:
            return result.astype(dtype)
        return result

    def longitudes(self, dtype: Optional[type] = None) -> Any:
        """Get the longitudes for the field.

        Parameters
        ----------
        dtype : Optional[type]
            The type to cast the longitudes to.

        Returns
        -------
        Any
            The longitudes for the field.
        """
        result = self._grid.longitudes
        if dtype is not None:
            return result.astype(dtype)
        return result

    def resolution(self) -> Optional[Any]:
        """Get the resolution for the field.

        Returns
        -------
        Optional[Any]
            The resolution for the field.
        """
        # TODO: implement resolution
        return None

    def mars_grid(self) -> Optional[Any]:
        """Get the MARS grid for the field.

        Returns
        -------
        Optional[Any]
            The MARS grid for the field.
        """
        # TODO: implement mars_grid
        return None

    def mars_area(self) -> Optional[Any]:
        """Get the MARS area for the field.

        Returns
        -------
        Optional[Any]
            The MARS area for the field.
        """
        # TODO: code me
        # return [self.north, self.west, self.south, self.east]
        return None

    def x(self, dtype: Optional[type] = None) -> None:
        """Get the x-coordinates for the field.

        Parameters
        ----------
        dtype : Optional[type]
            The type to cast the x-coordinates to.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError()

    def y(self, dtype: Optional[type] = None) -> None:
        """Get the y-coordinates for the field.

        Parameters
        ----------
        dtype : Optional[type]
            The type to cast the y-coordinates to.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError()

    def shape(self) -> Any:
        """Get the shape of the field.

        Returns
        -------
        Any
            The shape of the field.
        """
        return self._field.shape

    def projection(self) -> Projection:
        """Get the projection for the field.

        Returns
        -------
        Projection
            The projection for the field.
        """
        return Projection.from_cf_grid_mapping(**self._field.grid_mapping)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
from functools import cache
from functools import wraps
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type
from typing import Union
from unittest.mock import patch

import numpy as np
import zarr
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta

from anemoi.datasets import open_dataset

VALUES = 20


def mockup_open_zarr(func: Callable) -> Callable:
    """Decorator to mock the open_zarr function.

    Parameters
    ----------
    func : Callable
        Function to wrap.

    Returns
    -------
    Callable
        Wrapped function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        with patch("zarr.convenience.open", zarr_from_str):
            with patch("anemoi.datasets.data.stores.zarr_lookup", lambda name: name):
                return func(*args, **kwargs)

    return wrapper


@cache
def _(date: datetime.datetime, var: str, k: int = 0, e: int = 0, values: int = VALUES) -> np.ndarray:
    """Create a simple array of values based on the date and variable name, ensemble, grid, and other parameters.

    Parameters
    ----------
    date : datetime.datetime
        Date.
    var : str
        Variable name.
    k : int, optional
        Grid index, by default 0.
    e : int, optional
        Ensemble index, by default 0.
    values : int, optional
        Number of values, by default VALUES.

    Returns
    -------
    np.ndarray
        Array of values.
    """
    d = date.year * 10000 + date.month * 100 + date.day
    v = ord(var) - ord("a") + 1
    assert 0 <= k <= 9
    assert 0 <= e <= 9

    return np.array([d * 100 + v + k / 10.0 + w / 100.0 + e / 1000.0 for w in range(values)])


def create_zarr(
    vars: str = "abcd",
    start: int = 2021,
    end: int = 2021,
    field_shape: tuple = [4, 5],
    frequency: datetime.timedelta = datetime.timedelta(hours=6),
    resolution: str = "o96",
    k: int = 0,
    ensemble: Optional[int] = None,
    grids: Optional[int] = None,
    missing: bool = False,
) -> zarr.Group:
    """Create a Zarr dataset.

    Parameters
    ----------
    vars : str, optional
        Variable names, by default "abcd".
    start : int, optional
        Start year, by default 2021.
    end : int, optional
        End year, by default 2021.
    field_shape : tuple, optional
        Field shape of dataset, by default [4, 5].
    frequency : datetime.timedelta, optional
        Frequency, by default datetime.timedelta(hours=6).
    resolution : str, optional
        Resolution, by default "o96".
    k : int, optional
        Grid index, by default 0.
    ensemble : Optional[int], optional
        Number of ensembles, by default None.
    grids : Optional[int], optional
        Number of grids, by default None.
    missing : bool, optional
        Whether to include missing dates, by default False.

    Returns
    -------
    zarr.Group
        Zarr dataset.
    """
    root = zarr.group()
    assert isinstance(frequency, datetime.timedelta)

    dates = []
    date = datetime.datetime(start, 1, 1)
    while date.year <= end:
        dates.append(date)
        date += frequency

    dates = np.array(dates, dtype="datetime64")

    ensembles = ensemble if ensemble is not None else 1
    values = grids if grids is not None else VALUES

    data = np.zeros(shape=(len(dates), len(vars), ensembles, values))

    for i, date in enumerate(dates):
        for j, var in enumerate(vars):
            for e in range(ensembles):
                data[i, j, e] = _(date.astype(object), var, k, e, values)

    root.create_dataset(
        "data",
        data=data,
        dtype=data.dtype,
        chunks=data.shape,
        compressor=None,
    )
    root.create_dataset(
        "dates",
        data=dates,
        compressor=None,
    )
    root.create_dataset(
        "latitudes",
        data=np.array([x + values for x in range(values)]),
        compressor=None,
    )
    root.create_dataset(
        "longitudes",
        data=np.array([x + values for x in range(values)]),
        compressor=None,
    )

    root.attrs["frequency"] = frequency_to_string(frequency)
    root.attrs["resolution"] = resolution
    root.attrs["name_to_index"] = {k: i for i, k in enumerate(vars)}

    root.attrs["data_request"] = {"grid": 1, "area": "g", "param_level": {}}
    root.attrs["variables_metadata"] = {v: {} for v in vars}

    if missing:
        missing_dates = []

        last = None
        for date in [d.astype(object) for d in dates]:
            yyyymmdd = date.strftime("%Y%m")
            if yyyymmdd != last:
                last = yyyymmdd
                missing_dates.append(date)

        root.attrs["missing_dates"] = [d.isoformat() for d in missing_dates]

    root.create_dataset(
        "mean",
        data=np.mean(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "stdev",
        data=np.std(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "maximum",
        data=np.max(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "minimum",
        data=np.min(data, axis=0),
        compressor=None,
    )

    root.attrs["field_shape"] = field_shape

    return root


def zarr_from_str(name: str, mode: str) -> zarr.Group:
    """Create a Zarr dataset from a string.

    Parameters
    ----------
    name : str
        Dataset name.
    mode : str
        Mode.

    Returns
    -------
    zarr.Group
        Zarr dataset.
    """
    # Format: test-2021-2021-6h-o96-abcd-0

    args = dict(
        test="test",
        start=2021,
        end=2021,
        field_shape=[4, 5],
        frequency=6,
        resolution="o96",
        vars="abcd",
        k=0,
        ensemble=None,
        grids=None,
    )

    for name, bit in zip(args, name.split("-")):
        args[name] = bit

    args["field_shape"] = [int(i) for i in args["field_shape"].split(",")]

    print(args)

    return create_zarr(
        start=int(args["start"]),
        end=int(args["end"]),
        field_shape=args["field_shape"],
        frequency=frequency_to_timedelta(args["frequency"]),
        resolution=args["resolution"],
        vars=[x for x in args["vars"]],
        k=int(args["k"]),
        ensemble=int(args["ensemble"]) if args["ensemble"] is not None else None,
        grids=int(args["grids"]) if args["grids"] is not None else None,
        missing=args["test"] == "missing",
    )


class IndexTester:
    """Class to test indexing of datasets."""

    def __init__(self, ds: Any) -> None:
        """Initialize the IndexTester.

        Parameters
        ----------
        ds : Any
            Dataset.
        """
        self.ds = ds
        self.np = ds[:]  # Numpy array

        assert self.ds.shape == self.np.shape
        assert (self.ds == self.np).all()

    def __getitem__(self, index: Any) -> None:
        """Test indexing.

        Parameters
        ----------
        index : Any
            Index.
        """
        print("INDEX", type(self.ds), index)
        if self.ds[index] is None:
            assert False, (self.ds, index)

        if not (self.ds[index] == self.np[index]).all():
            # print("DS", self.ds[index])
            # print("NP", self.np[index])
            assert (self.ds[index] == self.np[index]).all()


def make_missing(x: Any) -> Any:
    """Mark data as missing.

    Parameters
    ----------
    x : Any
        Data.

    Returns
    -------
    Any
        Data with missing values.
    """
    if isinstance(x, tuple):
        return (make_missing(a) for a in x)
    if isinstance(x, list):
        return [make_missing(a) for a in x]
    if isinstance(x, dict):
        return {k: make_missing(v) for k, v in x.items()}
    if isinstance(x, str) and x.startswith("test-"):
        return x.replace("test-", "missing-")
    return x


class DatasetTester:
    """Class to test various dataset operations."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the DatasetTester.

        Parameters
        ----------
        *args : Any
            Arguments.
        **kwargs : Any
            Keyword arguments.
        """
        self.ds = open_dataset(*args, **kwargs)

        args, kwargs = make_missing((args, kwargs))

        print(f"ds={self.ds}")

    def run(
        self,
        *,
        expected_class: Type,
        expected_length: int,
        expected_shape: tuple,
        expected_variables: Union[str, list],
        expected_name_to_index: Union[str, dict],
        date_to_row: Callable,
        start_date: datetime.datetime,
        time_increment: datetime.timedelta,
        statistics_reference_dataset: Optional[Union[str, list]],
        statistics_reference_variables: Optional[Union[str, list]],
    ) -> None:
        """Run the dataset tests.

        Parameters
        ----------
        expected_class : Type
            Expected class.
        expected_length : int
            Expected length.
        expected_shape : tuple
            Expected shape.
        expected_variables : Union[str, list]
            Expected variables.
        expected_name_to_index : Union[str, dict]
            Expected name to index mapping.
        date_to_row : Callable
            Function to generate row data.
        start_date : datetime.datetime
            Start date.
        time_increment : datetime.timedelta
            Time increment.
        statistics_reference_dataset : Optional[Union[str, list]]
            Reference dataset for statistics.
        statistics_reference_variables : Optional[Union[str, list]]
            Reference variables for statistics.
        """
        if isinstance(expected_variables, str):
            expected_variables = [v for v in expected_variables]

        if isinstance(expected_name_to_index, str):
            expected_name_to_index = {v: i for i, v in enumerate(expected_name_to_index)}

        assert isinstance(self.ds, expected_class)
        assert len(self.ds) == expected_length
        assert len([row for row in self.ds]) == len(self.ds)
        assert self.ds.shape == expected_shape, (self.ds.shape, expected_shape)
        assert self.ds.variables == expected_variables

        assert set(self.ds.variables_metadata.keys()) == set(expected_variables)

        assert self.ds.name_to_index == expected_name_to_index
        assert self.ds.dates[0] == start_date
        assert self.ds.dates[1] - self.ds.dates[0] == time_increment

        dates = []
        date = start_date

        for row in self.ds:
            # print(f"{date=} {row.shape}")
            expect = date_to_row(date)
            assert (row == expect).all()
            dates.append(date)
            date += time_increment

        assert (self.ds.dates == np.array(dates, dtype="datetime64")).all()

        if statistics_reference_dataset is not None:
            self.same_stats(
                self.ds,
                open_dataset(statistics_reference_dataset),
                statistics_reference_variables,
            )

        self.indexing(self.ds)
        self.metadata(self.ds)

        self.ds.tree()

    def metadata(self, ds: Any) -> None:
        """Test metadata.

        Parameters
        ----------
        ds : Any
            Dataset.
        """
        metadata = ds.metadata()
        assert isinstance(metadata, dict)

    def same_stats(self, ds1: Any, ds2: Any, vars1: list, vars2: Optional[list] = None) -> None:
        """Compare statistics between two datasets.

        Parameters
        ----------
        ds1 : Any
            First dataset.
        ds2 : Any
            Second dataset.
        vars1 : list
            Variables in the first dataset.
        vars2 : Optional[list], optional
            Variables in the second dataset, by default None.
        """
        if vars2 is None:
            vars2 = vars1

        vars1 = [v for v in vars1]
        vars2 = [v for v in vars2]
        for v1, v2 in zip(vars1, vars2):
            idx1 = ds1.name_to_index[v1]
            idx2 = ds2.name_to_index[v2]
            assert (ds1.statistics["mean"][idx1] == ds2.statistics["mean"][idx2]).all()
            assert (ds1.statistics["stdev"][idx1] == ds2.statistics["stdev"][idx2]).all()
            assert (ds1.statistics["maximum"][idx1] == ds2.statistics["maximum"][idx2]).all()
            assert (ds1.statistics["minimum"][idx1] == ds2.statistics["minimum"][idx2]).all()

    def indexing(self, ds: Any) -> None:
        """Test indexing.

        Parameters
        ----------
        ds : Any
            Dataset.
        """
        t = IndexTester(ds)

        print("INDEXING", ds.shape)

        t[0:10, :, 0]
        t[:, 0:3, 0]
        # t[:, :, 0]
        t[0:10, 0:3, 0]
        t[:, :, :]

        if ds.shape[1] > 2:  # Variable dimension
            t[:, (1, 2), :]
            t[:, (1, 2)]

        t[0]
        t[0, :]
        t[0, 0, :]
        t[0, 0, 0, :]

        if ds.shape[2] > 1:  # Ensemble dimension
            t[0:10, :, (0, 1)]

        for i in range(3):
            t[i]
            start = 5 * i
            end = len(ds) - 5 * i
            step = len(ds) // 10

            t[start:end:step]
            t[start:end]
            t[start:]
            t[:end]
            t[::step]


@mockup_open_zarr
def test_trim_edge_simple() -> None:
    """Test trimming the edges of a dataset."""
    test = DatasetTester(
        "test-2021-2021-15,14-6h-o96-abcd",
        trim_edge=(2, 3, 4, 5),
    )

    expected_field_shape = (10, 5)
    assert test.ds.field_shape == expected_field_shape, test.ds.field_shape
    assert test.ds.shape == (365 * 4, 4, 1, np.prod(expected_field_shape)), test.ds.shape


@mockup_open_zarr
def test_trim_edge_zeros() -> None:
    """Test trimming the edges of a dataset when edges are 0"""
    for dim in range(2):
        trim_edge = [0, 0, 0, 0]
        trim_edge[dim] = 1
        test = DatasetTester(
            "test-2021-2021-15,14-6h-o96-abcd",
            trim_edge=trim_edge,
        )

        expected_field_shape = (14, 14)
        assert test.ds.field_shape == expected_field_shape, test.ds.field_shape
        assert test.ds.shape == (365 * 4, 4, 1, np.prod(expected_field_shape)), test.ds.shape

    for dim in range(2, 4):
        trim_edge = [0, 0, 0, 0]
        trim_edge[dim] = 1
        test = DatasetTester(
            "test-2021-2021-15,14-6h-o96-abcd",
            trim_edge=trim_edge,
        )

        expected_field_shape = (15, 13)
        assert test.ds.field_shape == expected_field_shape, test.ds.field_shape
        assert test.ds.shape == (365 * 4, 4, 1, np.prod(expected_field_shape)), test.ds.shape


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd
import xarray as xr
from earthkit.data.core.fieldlist import MultiFieldList

from anemoi.datasets.create.sources.patterns import iterate_patterns
from anemoi.datasets.data.stores import name_to_zarr_store

from ..legacy import legacy_source
from .fieldlist import XarrayFieldList

LOG = logging.getLogger(__name__)


def check(what: str, ds: xr.Dataset, paths: List[str], **kwargs: Any) -> None:
    """Checks if the dataset has the expected number of fields.

    Parameters
    ----------
    what : str
        Description of what is being checked.
    ds : xr.Dataset
        The dataset to check.
    paths : List[str]
        List of paths.
    **kwargs : Any
        Additional keyword arguments.
    """
    count = 1
    for k, v in kwargs.items():
        if isinstance(v, (tuple, list)):
            count *= len(v)

    if len(ds) != count:
        raise ValueError(f"Expected {count} fields, got {len(ds)} (kwargs={kwargs}, {what}s={paths})")


def load_one(
    emoji: str,
    context: Any,
    dates: List[str],
    dataset: Union[str, xr.Dataset],
    *,
    options: Optional[Dict[str, Any]] = None,
    flavour: Optional[str] = None,
    patch: Optional[Any] = None,
    **kwargs: Any,
) -> ekd.FieldList:
    """Loads a single dataset.

    Parameters
    ----------
    emoji : str
        Emoji for tracing.
    context : Any
        Context object.
    dates : List[str]
        List of dates.
    dataset : Union[str, xr.Dataset]
        The dataset to load.
    options : Dict[str, Any], optional
        Additional options for loading the dataset.
    flavour : Optional[str], optional
        Flavour of the dataset.
    patch : Optional[Any], optional
        Patch for the dataset.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    MultiFieldList
        The loaded dataset.
    """

    """
    We manage the S3 client ourselves, bypassing fsspec and s3fs layers, because sometimes something on the stack
    zarr/fsspec/s3fs/boto3 (?) seem to flags files as missing when they actually are not (maybe when S3 reports some sort of
    connection error). In that case,  Zarr will silently fill the chunks that could not be downloaded with NaNs.
    See https://github.com/pydata/xarray/issues/8842

    We have seen this bug triggered when we run many clients in parallel, for example, when we create a new dataset using `xarray-zarr`.
    """

    if options is None:
        options = {}

    context.trace(emoji, dataset, options, kwargs)

    if isinstance(dataset, str) and ".zarr" in dataset:
        data = xr.open_zarr(name_to_zarr_store(dataset), **options)
    elif "planetarycomputer" in dataset:
        store = name_to_zarr_store(dataset)
        if "store" in store:
            data = xr.open_zarr(**store)
        if "filename_or_obj" in store:
            data = xr.open_dataset(**store)
    else:
        data = xr.open_dataset(dataset, **options)

    fs = XarrayFieldList.from_xarray(data, flavour=flavour, patch=patch)

    if len(dates) == 0:
        result = fs.sel(**kwargs)
    else:
        print("dates", dates, kwargs)
        result = MultiFieldList([fs.sel(valid_datetime=date, **kwargs) for date in dates])

    if len(result) == 0:
        LOG.warning(f"No data found for {dataset} and dates {dates} and {kwargs}")
        LOG.warning(f"Options: {options}")

        for i, k in enumerate(fs):
            a = ["valid_datetime", k.metadata("valid_datetime", default=None)]
            for n in kwargs.keys():
                a.extend([n, k.metadata(n, default=None)])
            print([str(x) for x in a])

            if i > 16:
                break

        # LOG.warning(data)

    return result


def load_many(emoji: str, context: Any, dates: List[datetime.datetime], pattern: str, **kwargs: Any) -> ekd.FieldList:
    """Loads multiple datasets.

    Parameters
    ----------
    emoji : str
        Emoji for tracing.
    context : Any
        Context object.
    dates : List[str]
        List of dates.
    pattern : str
        Pattern for loading datasets.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    MultiFieldList
        The loaded datasets.
    """
    result = []

    for path, dates in iterate_patterns(pattern, dates, **kwargs):
        result.append(load_one(emoji, context, dates, path, **kwargs))

    return MultiFieldList(result)


@legacy_source("xarray")
def execute(context: Any, dates: List[str], url: str, *args: Any, **kwargs: Any) -> ekd.FieldList:
    """Executes the loading of datasets.

    Parameters
    ----------
    context : Any
        Context object.
    dates : List[str]
        List of dates.
    url : str
        URL pattern for loading datasets.
    *args : Any
        Additional arguments.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    ekd.FieldList
        The loaded datasets.
    """
    return load_many("🌐", context, dates, url, *args, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any
from typing import Dict
from typing import List

import earthkit.data as ekd
import numpy as np
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from earthkit.meteo import thermo

from .legacy import legacy_filter
from .single_level_specific_humidity_to_relative_humidity import AutoDict
from .single_level_specific_humidity_to_relative_humidity import pressure_at_height_level


@legacy_filter(__file__)
def execute(
    context: Any,
    input: List[Any],
    height: float,
    t: str,
    rh: str,
    sp: str,
    new_name: str = "2q",
    **kwargs: Dict[str, Any],
) -> ekd.FieldList:
    """Convert the single (height) level relative humidity to specific humidity.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : List[Any]
        List of input fields.
    height : float
        The height level.
    t : str
        Temperature parameter name.
    rh : str
        Relative humidity parameter name.
    sp : str
        Surface pressure parameter name.
    new_name : str, optional
        The new name for the specific humidity field, by default "2q".
    **kwargs : Dict[str, Any]
        Additional keyword arguments.

    Returns
    -------
    ekd.FieldList
        The resulting field list with specific humidity fields.

    Raises
    ------
    KeyError
        If mandatory keys are missing.
    ValueError
        If there are duplicate fields or missing fields.
    """
    result = []

    MANDATORY_KEYS = ["A", "B"]
    OPTIONAL_KEYS = ["t_ml", "q_ml"]
    MISSING_KEYS = []
    DEFAULTS = dict(t_ml="t", q_ml="q")

    for key in OPTIONAL_KEYS:
        if key not in kwargs:
            print(f"key {key} not found in yaml-file, using default key: {DEFAULTS[key]}")
            kwargs[key] = DEFAULTS[key]

    for key in MANDATORY_KEYS:
        if key not in kwargs:
            MISSING_KEYS.append(key)

    if MISSING_KEYS:
        raise KeyError(f"Following keys are missing: {', '.join(MISSING_KEYS)}")

    single_level_params = (t, rh, sp)
    model_level_params = (kwargs["t_ml"], kwargs["q_ml"])

    needed_fields = AutoDict()

    # Gather all necessary fields
    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        # check single level parameters
        if param in single_level_params:
            levtype = key.pop("levtype")
            key = tuple(key.items())

            if param in needed_fields[key][levtype]:
                raise ValueError(f"Duplicate single level field {param} for {key}")

            needed_fields[key][levtype][param] = f
            if param == rh:
                if kwargs.get("keep_rh", False):
                    result.append(f)
            else:
                result.append(f)

        # check model level parameters
        elif param in model_level_params:
            levtype = key.pop("levtype")
            levelist = key.pop("levelist")
            key = tuple(key.items())

            if param in needed_fields[key][levtype][levelist]:
                raise ValueError(f"Duplicate model level field {param} for {key} at level {levelist}")

            needed_fields[key][levtype][levelist][param] = f

        # all other parameters
        else:
            result.append(f)

    for _, values in needed_fields.items():
        # some checks
        if len(values["sfc"]) != 3:
            raise ValueError("Missing surface fields")

        rh_sl = values["sfc"][rh].to_numpy(flatten=True)
        t_sl = values["sfc"][t].to_numpy(flatten=True)
        sp_sl = values["sfc"][sp].to_numpy(flatten=True)

        nlevels = len(kwargs["A"]) - 1
        if len(values["ml"]) != nlevels:
            raise ValueError("Missing model levels")

        for key in values["ml"].keys():
            if len(values["ml"][key]) != 2:
                raise ValueError(f"Missing field on level {key}")

        # create 3D arrays for upper air fields
        levels = list(values["ml"].keys())
        levels.sort()
        t_ml = []
        q_ml = []
        for level in levels:
            t_ml.append(values["ml"][level][kwargs["t_ml"]].to_numpy(flatten=True))
            q_ml.append(values["ml"][level][kwargs["q_ml"]].to_numpy(flatten=True))

        t_ml = np.stack(t_ml)
        q_ml = np.stack(q_ml)

        # actual conversion from rh --> q_v
        p_sl = pressure_at_height_level(height, q_ml, t_ml, sp_sl, np.array(kwargs["A"]), np.array(kwargs["B"]))
        q_sl = thermo.specific_humidity_from_relative_humidity(t_sl, rh_sl, p_sl)

        result.append(new_field_from_numpy(values["sfc"][rh], q_sl, param=new_name))

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import textwrap
from typing import Any
from typing import List
from typing import Tuple
from typing import Union

from anemoi.utils.humanize import plural

from .trace import step
from .trace import trace

LOG = logging.getLogger(__name__)


class Context:
    """Class to handle the build context in the dataset creation process."""

    def __init__(self) -> None:
        """Initializes a Context instance."""
        # used_references is a set of reference paths that will be needed
        self.used_references = set()
        # results is a dictionary of reference path -> obj
        self.results = {}

    def will_need_reference(self, key: Union[List, Tuple]) -> None:
        """Marks a reference as needed.

        Parameters
        ----------
        key : Union[List, Tuple]
            The reference key.
        """
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        self.used_references.add(key)

    def notify_result(self, key: Union[List, Tuple], result: Any) -> None:
        """Notifies that a result is available for a reference.

        Parameters
        ----------
        key : Union[List, Tuple]
            The reference key.
        result : Any
            The result object.
        """
        trace(
            "🎯",
            step(key),
            "notify result",
            textwrap.shorten(repr(result).replace(",", ", "), width=40),
            plural(len(result), "field"),
        )
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        if key in self.used_references:
            if key in self.results:
                raise ValueError(f"Duplicate result {key}")
            self.results[key] = result

    def get_result(self, key: Union[List, Tuple]) -> Any:
        """Retrieves the result for a given reference.

        Parameters
        ----------
        key : Union[List, Tuple]
            The reference key.

        Returns
        -------
        Any
            The result for the given reference.
        """
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        if key in self.results:
            return self.results[key]
        all_keys = sorted(list(self.results.keys()))
        raise ValueError(f"Cannot find result {key} in {all_keys}")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

import numpy as np
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .forwards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class Join(Combined):
    """Join the datasets along the variables axis."""

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Check the compatibility of two datasets.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=1)

    def check_same_variables(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the datasets have the same variables.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # Turned off because we are joining along the variables axis
        pass

    def __len__(self) -> int:
        """Get the length of the joined dataset.

        Returns
        -------
        int
            The length of the joined dataset.
        """
        return len(self.datasets[0])

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get the data for a tuple index.

        Parameters
        ----------
        index : TupleIndex
            The tuple index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data for the tuple index.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 1, slice(None))

        # TODO: optimize if index does not access all datasets, so we don't load chunks we don't need
        result = [d[index] for d in self.datasets]

        result = np.concatenate(result, axis=1)
        return apply_index_to_slices_changes(result[:, previous], changes)

    @debug_indexing
    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Get the data for a slice.

        Parameters
        ----------
        s : slice
            The slice to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data for the slice.
        """
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get the data at the specified index.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data at the specified index.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        return np.concatenate([d[n] for d in self.datasets])

    @cached_property
    def shape(self) -> Shape:
        """Get the shape of the joined dataset."""
        cols = sum(d.shape[1] for d in self.datasets)
        return (len(self), cols) + self.datasets[0].shape[2:]

    def _overlay(self) -> Dataset:
        """Overlay the datasets.

        Returns
        -------
        Dataset
            The overlaid dataset.
        """
        indices = {}
        i = 0
        for d in self.datasets:
            for v in d.variables:
                indices[v] = i
                i += 1

        if len(indices) == i:
            # No overlay
            return self

        variables = [v[1:-1] for v in self.variables if v[0] == "(" and v[-1] == ")"]
        indices = list(indices.values())

        i = 0
        for d in self.datasets:
            ok = False
            for v in d.variables:
                if i in indices:
                    ok = True
                i += 1
            if not ok:
                LOG.warning("Dataset %r completely overridden.", d)

        from .select import Select

        return Select(self, indices, {"overlay": variables})

    @cached_property
    def variables(self) -> List[str]:
        """Get the variables of the joined dataset."""
        seen = set()
        result: List[str] = []
        for d in reversed(self.datasets):
            for v in reversed(d.variables):
                while v in seen:
                    v = f"({v})"
                seen.add(v)
                result.insert(0, v)

        return result

    @property
    def variables_metadata(self) -> Dict[str, Any]:
        """Get the metadata of the variables."""
        result = {}
        variables = [v for v in self.variables if not (v.startswith("(") and v.endswith(")"))]

        for d in self.datasets:
            md = d.variables_metadata
            for v in variables:
                if v in md:
                    result[v] = md[v]

        if len(result) != len(variables):
            LOG.error("Some variables are missing metadata.")
            for v in variables:
                if v not in result:
                    LOG.error("Missing metadata for %r.", v)

        return result

    @cached_property
    def name_to_index(self) -> Dict[str, int]:
        """Get the mapping of variable names to indices."""
        return {k: i for i, k in enumerate(self.variables)}

    @property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Get the statistics of the joined dataset."""
        return {
            k: np.concatenate([d.statistics[k] for d in self.datasets], axis=0) for k in self.datasets[0].statistics
        }

    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Get the statistics tendencies of the joined dataset.

        Parameters
        ----------
        delta : Optional[datetime.timedelta]
            The time delta for the tendencies.

        Returns
        -------
        Dict[str, NDArray[Any]]
            The statistics tendencies of the joined dataset.
        """
        if delta is None:
            delta = self.frequency
        return {
            k: np.concatenate([d.statistics_tendencies(delta)[k] for d in self.datasets], axis=0)
            for k in self.datasets[0].statistics_tendencies(delta)
        }

    def source(self, index: int) -> Source:
        """Get the source of the data at the specified index.

        Parameters
        ----------
        index : int
            The index to retrieve the source from.

        Returns
        -------
        Source
            The source of the data at the specified index.
        """
        i = index
        for dataset in self.datasets:
            if i < dataset.shape[1]:
                return Source(self, index, dataset.source(i))
            i -= dataset.shape[1]
        assert False

    @cached_property
    def missing(self) -> Set[int]:
        """Get the missing data indices."""
        result: Set[int] = set()
        for d in self.datasets:
            result = result | d.missing
        return result

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation of the dataset.
        """
        return Node(self, [d.tree() for d in self.datasets])

    def forwards_subclass_metadata_specific(self) -> dict[str, Any]:
        """Get the metadata specific to the forwards subclass.

        Returns
        -------
        dict[str, Any]
            The metadata specific to the forwards subclass.
        """
        return {}


def join_factory(args: tuple, kwargs: dict) -> Dataset:
    """Create a joined dataset.

    Parameters
    ----------
    args : tuple
        The positional arguments.
    kwargs : dict
        The keyword arguments.

    Returns
    -------
    Dataset
        The joined dataset.
    """
    datasets = kwargs.pop("join")
    assert isinstance(datasets, (list, tuple))
    assert len(args) == 0

    assert isinstance(datasets, (list, tuple))

    datasets = [_open(e) for e in datasets]

    if len(datasets) == 1:
        return datasets[0]._subset(**kwargs)

    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Join(datasets)._overlay()._subset(**kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from earthkit.data import FieldList

from ...dates.groups import GroupOfDates
from .action import Action
from .action import action_factory
from .misc import _tidy
from .result import Result

LOG = logging.getLogger(__name__)


class DataSourcesAction(Action):
    """Class to handle data sources actions in the dataset creation process."""

    def __init__(
        self,
        context: object,
        action_path: List[str],
        sources: Union[Dict[str, Any], List[Dict[str, Any]]],
        input: Dict[str, Any],
    ) -> None:
        """Initializes a DataSourcesAction instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : List[str]
            The action path.
        sources : Union[Dict[str, Any], List[Dict[str, Any]]]
            The sources configuration.
        input : Dict[str, Any]
            The input configuration.
        """
        super().__init__(context, ["data_sources"], *sources)
        if isinstance(sources, dict):
            configs = [(str(k), c) for k, c in sources.items()]
        elif isinstance(sources, list):
            configs = [(str(i), c) for i, c in enumerate(sources)]
        else:
            raise ValueError(f"Invalid data_sources, expecting list or dict, got {type(sources)}: {sources}")

        self.sources = [action_factory(config, context, ["data_sources"] + [a_path]) for a_path, config in configs]
        self.input = action_factory(input, context, ["input"])

    def select(self, group_of_dates: GroupOfDates) -> "DataSourcesResult":
        """Selects the data sources result for the given group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            The group of dates.

        Returns
        -------
        DataSourcesResult
            The data sources result.
        """
        sources_results = [a.select(group_of_dates) for a in self.sources]
        return DataSourcesResult(
            self.context,
            self.action_path,
            group_of_dates,
            self.input.select(group_of_dates),
            sources_results,
        )

    def __repr__(self) -> str:
        """Returns a string representation of the DataSourcesAction instance."""
        content = "\n".join([str(i) for i in self.sources])
        return self._repr(content)


class DataSourcesResult(Result):
    """Class to represent the result of data sources actions in the dataset creation process."""

    def __init__(
        self,
        context: object,
        action_path: List[str],
        dates: object,
        input_result: Result,
        sources_results: List[Result],
    ) -> None:
        """Initializes a DataSourcesResult instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : List[str]
            The action path.
        dates : object
            The dates object.
        input_result : Result
            The input result.
        sources_results : List[Result]
            The list of sources results.
        """
        super().__init__(context, action_path, dates)
        # result is the main input result
        self.input_result = input_result
        # sources_results is the list of the sources_results
        self.sources_results = sources_results

    @cached_property
    def datasource(self) -> FieldList:
        """Returns the combined datasource from all sources."""
        for i in self.sources_results:
            # for each result trigger the datasource to be computed
            # and saved in context
            self.context.notify_result(i.action_path[:-1], i.datasource)
        # then return the input result
        # which can use the datasources of the included results
        return _tidy(self.input_result.datasource)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import warnings
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import earthkit.data as ekd
import numpy as np
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.output import new_grib_output
from numpy.typing import NDArray

from anemoi.datasets.create.utils import to_datetime_list

from .legacy import legacy_source
from .mars import mars

LOG = logging.getLogger(__name__)


def _member(field: Any) -> int:
    """Retrieves the member number from the field metadata.

    Parameters
    ----------
    field : Any
        The field from which to retrieve the member number.

    Returns
    -------
    int
        The member number.
    """
    # Bug in eccodes has number=0 randomly
    number = field.metadata("number", default=0)
    if number is None:
        number = 0
    return number


class Accumulation:
    """Class to handle data accumulation for a specific parameter, date, time, and member."""

    buggy_steps: bool = False

    def __init__(
        self,
        out: Any,
        /,
        param: str,
        date: int,
        time: int,
        number: int,
        step: List[int],
        frequency: int,
        **kwargs: Any,
    ) -> None:
        """Initializes an Accumulation instance.

        Parameters
        ----------
        out : Any
            Output object for writing data.
        param : str
            Parameter name.
        date : int
            Date of the accumulation.
        time : int
            Time of the accumulation.
        number : int
            Member number.
        step : List[int]
            List of steps.
        frequency : int
            Frequency of accumulation.
        **kwargs : Any
            Additional keyword arguments.
        """
        self.out = out
        self.param = param
        self.date = date
        self.time = time
        self.steps = step
        self.number = number
        self.values: Optional[NDArray[None]] = None
        self.seen = set()
        self.startStep: Optional[int] = None
        self.endStep: Optional[int] = None
        self.done = False
        self.frequency = frequency
        self._check = None

    @property
    def key(self) -> Tuple[str, int, int, List[int], int]:
        """Returns the key for the accumulation."""
        return (self.param, self.date, self.time, self.steps, self.number)

    def check(self, field: Any) -> None:
        """Checks the field metadata against the accumulation parameters.

        Parameters
        ----------
        field : Any
            The field to check.
        """
        if self._check is None:
            self._check = field.metadata(namespace="mars")

            assert self.param == field.metadata("param"), (
                self.param,
                field.metadata("param"),
            )
            assert self.date == field.metadata("date"), (
                self.date,
                field.metadata("date"),
            )
            assert self.time == field.metadata("time"), (
                self.time,
                field.metadata("time"),
            )
            assert self.number == _member(field), (self.number, _member(field))

            return

        mars = field.metadata(namespace="mars")
        keys1 = sorted(self._check.keys())
        keys2 = sorted(mars.keys())

        assert keys1 == keys2, (keys1, keys2)

        for k in keys1:
            if k not in ("step",):
                assert self._check[k] == mars[k], (k, self._check[k], mars[k])

    def write(self, template: Any) -> None:
        """Writes the accumulated values to the output.

        Parameters
        ----------
        template : Any
            Template for writing the output.
        """
        assert self.startStep != self.endStep, (self.startStep, self.endStep)
        if np.all(self.values < 0):
            LOG.warning(
                f"Negative values when computing accumutation for {self.param} ({self.date} {self.time}): min={np.amin(self.values)} max={np.amax(self.values)}"
            )

        self.out.write(
            self.values,
            template=template,
            stepType="accum",
            startStep=self.startStep,
            endStep=self.endStep,
            check_nans=True,
        )
        self.values = None
        self.done = True

    def add(self, field: Any, values: NDArray[Any]) -> None:
        """Adds values to the accumulation.

        Parameters
        ----------
        field : Any
            The field containing the values.
        values : np.ndarray
            The values to add.
        """
        self.check(field)

        step = field.metadata("step")
        if step not in self.steps:
            return

        if not np.all(values >= 0):
            warnings.warn(f"Negative values for {field}: {np.amin(values)} {np.amax(values)}")

        assert not self.done, (self.key, step)
        assert step not in self.seen, (self.key, step)

        startStep = field.metadata("startStep")
        endStep = field.metadata("endStep")

        if self.buggy_steps and startStep == endStep:
            startStep = 0

        assert step == endStep, (startStep, endStep, step)

        self.compute(values, startStep, endStep)

        self.seen.add(step)

        if len(self.seen) == len(self.steps):
            self.write(template=field)

    @classmethod
    def mars_date_time_steps(
        cls,
        dates: List[datetime.datetime],
        step1: int,
        step2: int,
        frequency: Optional[int],
        base_times: List[int],
        adjust_step: bool,
    ) -> Generator[Tuple[int, int, Tuple[int, ...]], None, None]:
        """Generates MARS date-time steps.

        Parameters
        ----------
        dates : List[datetime.datetime]
            List of dates.
        step1 : int
            First step.
        step2 : int
            Second step.
        frequency : Optional[int]
            Frequency of accumulation.
        base_times : List[int]
            List of base times.
        adjust_step : bool
            Whether to adjust the step.

        Returns
        -------
        Generator[Tuple[int, int, Tuple[int, ...]], None, None]
            A generator of MARS date-time steps.
        """
        # assert step1 > 0, (step1, step2, frequency)

        for valid_date in dates:
            base_date = valid_date - datetime.timedelta(hours=step2)
            add_step = 0
            if base_date.hour not in base_times:
                if not adjust_step:
                    raise ValueError(
                        f"Cannot find a base time in {base_times} that validates on {valid_date.isoformat()} for step={step2}"
                    )

                while base_date.hour not in base_times:
                    # print(f'{base_date=}, {base_times=}, {add_step=} {frequency=}')
                    base_date -= datetime.timedelta(hours=1)
                    add_step += 1

            yield cls._mars_date_time_step(base_date, step1, step2, add_step, frequency)

    def __repr__(self) -> str:
        """Returns a string representation of the Accumulation instance.

        Returns
        -------
        str
            String representation of the Accumulation instance.
        """
        return f"{self.__class__.__name__}({self.key})"


class AccumulationFromStart(Accumulation):
    """Class to handle data accumulation from the start of the forecast."""

    buggy_steps = True

    def compute(self, values: NDArray[Any], startStep: int, endStep: int) -> None:
        """Computes the accumulation from the start.

        Parameters
        ----------
        values : np.ndarray
            The values to accumulate.
        startStep : int
            The start step.
        endStep : int
            The end step.
        """
        assert startStep == 0, startStep

        if self.values is None:

            self.values = np.copy(values)
            self.startStep = 0
            self.endStep = endStep

        else:
            assert endStep != self.endStep, (self.endStep, endStep)

            if endStep > self.endStep:
                # assert endStep - self.endStep == self.stepping, (self.endStep, endStep, self.stepping)
                self.values = values - self.values
                self.startStep = self.endStep
                self.endStep = endStep
            else:
                # assert self.endStep - endStep == self.stepping, (self.endStep, endStep, self.stepping)
                self.values = self.values - values
                self.startStep = endStep

            if not np.all(self.values >= 0):
                warnings.warn(f"Negative values for {self.param}: {np.amin(self.values)} {np.amax(self.values)}")
                self.values = np.maximum(self.values, 0)

    @classmethod
    def _mars_date_time_step(
        cls, base_date: datetime.datetime, step1: int, step2: int, add_step: int, frequency: Optional[int]
    ) -> Tuple[int, int, Tuple[int, ...]]:
        """Generates a MARS date-time step.

        Parameters
        ----------
        base_date : datetime.datetime
            The base date.
        step1 : int
            First step.
        step2 : int
            Second step.
        add_step : int
            Additional step.
        frequency : Optional[int]
            Frequency of accumulation.

        Returns
        -------
        Tuple[int, int, Tuple[int, ...]]
            A tuple representing the MARS date-time step.
        """
        assert not frequency, frequency

        steps = (step1 + add_step, step2 + add_step)
        if steps[0] == 0:
            steps = (steps[1],)

        return (
            base_date.year * 10000 + base_date.month * 100 + base_date.day,
            base_date.hour * 100 + base_date.minute,
            steps,
        )


class AccumulationFromLastStep(Accumulation):
    """Class to handle data accumulation from the last step of the forecast."""

    buggy_steps = False

    def compute(self, values: NDArray[Any], startStep: int, endStep: int) -> None:
        """Computes the accumulation from the last step.

        Parameters
        ----------
        values : np.ndarray
            The values to accumulate.
        startStep : int
            The start step.
        endStep : int
            The end step.
        """
        assert endStep - startStep == self.frequency, (
            startStep,
            endStep,
            self.frequency,
        )

        if self.startStep is None:
            self.startStep = startStep
        else:
            self.startStep = min(self.startStep, startStep)

        if self.endStep is None:
            self.endStep = endStep
        else:
            self.endStep = max(self.endStep, endStep)

        if self.values is None:
            self.values = np.zeros_like(values)

        self.values += values

    @classmethod
    def _mars_date_time_step(
        cls, base_date: datetime.datetime, step1: int, step2: int, add_step: int, frequency: int
    ) -> Tuple[int, int, Tuple[int, ...]]:
        """Generates a MARS date-time step.

        Parameters
        ----------
        base_date : datetime.datetime
            The base date.
        step1 : int
            First step.
        step2 : int
            Second step.
        add_step : int
            Additional step.
        frequency : int
            Frequency of accumulation.

        Returns
        -------
        Tuple[int, int, Tuple[int, ...]]
            A tuple representing the MARS date-time step.
        """
        assert frequency > 0, frequency
        # assert step1 > 0, (step1, step2, frequency, add_step, base_date)

        steps = []
        for step in range(step1 + frequency, step2 + frequency, frequency):
            steps.append(step + add_step)
        return (
            base_date.year * 10000 + base_date.month * 100 + base_date.day,
            base_date.hour * 100 + base_date.minute,
            tuple(steps),
        )


def _identity(x: Any) -> Any:
    """Identity function that returns the input as is.

    Parameters
    ----------
    x : Any
        Input value.

    Returns
    -------
    Any
        The input value.
    """
    return x


def _compute_accumulations(
    context: Any,
    dates: List[datetime.datetime],
    request: Dict[str, Any],
    user_accumulation_period: Union[int, Tuple[int, int]] = 6,
    data_accumulation_period: Optional[int] = None,
    patch: Any = _identity,
    base_times: Optional[List[int]] = None,
    use_cdsapi_dataset: Optional[str] = None,
) -> Any:
    """Computes accumulations based on the provided parameters.

    Parameters
    ----------
    context : Any
        Context for the computation.
    dates : List[datetime.datetime]
        List of dates.
    request : Dict[str, Any]
        Request parameters.
    user_accumulation_period : Union[int, Tuple[int, int]], optional
        User-defined accumulation period. Defaults to 6.
    data_accumulation_period : Optional[int], optional
        Data accumulation period. Defaults to None.
    patch : Any, optional
        Patch function. Defaults to _identity.
    base_times : Optional[List[int]], optional
        List of base times. Defaults to None.
    use_cdsapi_dataset : Optional[str], optional
        CDSAPI dataset to use. Defaults to None.

    Returns
    -------
    Any
        The computed accumulations.
    """
    adjust_step = isinstance(user_accumulation_period, int)

    if not isinstance(user_accumulation_period, (list, tuple)):
        user_accumulation_period = (0, user_accumulation_period)

    assert len(user_accumulation_period) == 2, user_accumulation_period
    step1, step2 = user_accumulation_period
    assert step1 < step2, user_accumulation_period

    if base_times is None:
        base_times = [0, 6, 12, 18]

    base_times = [t // 100 if t > 100 else t for t in base_times]

    AccumulationClass = AccumulationFromStart if data_accumulation_period in (0, None) else AccumulationFromLastStep

    mars_date_time_steps = AccumulationClass.mars_date_time_steps(
        dates,
        step1,
        step2,
        data_accumulation_period,
        base_times,
        adjust_step,
    )

    request = deepcopy(request)

    param = request["param"]
    if not isinstance(param, (list, tuple)):
        param = [param]

    number = request.get("number", [0])
    assert isinstance(number, (list, tuple))

    frequency = data_accumulation_period

    type_ = request.get("type", "an")
    if type_ == "an":
        type_ = "fc"

    request.update({"type": type_, "levtype": "sfc"})

    tmp = temp_file()
    path = tmp.path
    out = new_grib_output(path)

    requests = []

    accumulations = {}

    for date, time, steps in mars_date_time_steps:
        for p in param:
            for n in number:
                r = dict(request, param=p, date=date, time=time, step=sorted(steps), number=n)

                requests.append(patch(r))

    ds = mars(
        context, dates, *requests, request_already_using_valid_datetime=True, use_cdsapi_dataset=use_cdsapi_dataset
    )

    accumulations = {}
    for a in [AccumulationClass(out, frequency=frequency, **r) for r in requests]:
        for s in a.steps:
            key = (a.param, a.date, a.time, s, a.number)
            accumulations.setdefault(key, []).append(a)

    for field in ds:
        key = (
            field.metadata("param"),
            field.metadata("date"),
            field.metadata("time"),
            field.metadata("step"),
            _member(field),
        )
        values = field.values  # optimisation
        if key not in accumulations:
            raise ValueError(f"Key not found: {key}. Is it an accumulation field?")

        for a in accumulations[key]:
            a.add(field, values)

    for acc in accumulations.values():
        for a in acc:
            assert a.done, (a.key, a.seen, a.steps)

    out.close()

    ds = ekd.from_source("file", path)

    assert len(ds) / len(param) / len(number) == len(dates), (
        len(ds),
        len(param),
        len(dates),
    )
    ds._tmp = tmp

    return ds


def _to_list(x: Union[List[Any], Tuple[Any], Any]) -> List[Any]:
    """Converts the input to a list if it is not already a list or tuple.

    Parameters
    ----------
    x : Union[List[Any], Tuple[Any], Any]
        Input value.

    Returns
    -------
    List[Any]
        The input value as a list.
    """
    if isinstance(x, (list, tuple)):
        return x
    return [x]


def _scda(request: Dict[str, Any]) -> Dict[str, Any]:
    """Modifies the request stream based on the time.

    Parameters
    ----------
    request : Dict[str, Any]
        Request parameters.

    Returns
    -------
    Dict[str, Any]
        The modified request parameters.
    """
    if request["time"] in (6, 18, 600, 1800):
        request["stream"] = "scda"
    else:
        request["stream"] = "oper"
    return request


@legacy_source(__file__)
def accumulations(
    context: Any, dates: List[datetime.datetime], use_cdsapi_dataset: Optional[str] = None, **request: Any
) -> Any:
    """Computes accumulations based on the provided context, dates, and request parameters.

    Parameters
    ----------
    context : Any
        Context for the computation.
    dates : List[datetime.datetime]
        List of dates.
    use_cdsapi_dataset : Optional[str], optional
        CDSAPI dataset to use. Defaults to None.
    **request : Any
        Additional request parameters.

    Returns
    -------
    Any
        The computed accumulations.
    """

    if (
        request.get("class") == "ea"
        and request.get("stream", "oper") == "oper"
        and request.get("accumulation_period") == 24
    ):
        from .accumulations2 import accumulations as accumulations2

        LOG.warning(
            "🧪️ Experimental features: Using accumulations2, because class=ea stream=oper and accumulation_period=24"
        )
        return accumulations2(context, dates, **request)

    _to_list(request["param"])
    class_ = request.get("class", "od")
    stream = request.get("stream", "oper")

    user_accumulation_period = request.pop("accumulation_period", 6)

    # If `data_accumulation_period` is not set, this means that the accumulations are from the start
    # of the forecast.

    KWARGS = {
        ("od", "oper"): dict(patch=_scda),
        ("od", "elda"): dict(base_times=(6, 18)),
        ("od", "enfo"): dict(base_times=(0, 6, 12, 18)),
        ("ea", "oper"): dict(data_accumulation_period=1, base_times=(6, 18)),
        ("ea", "enda"): dict(data_accumulation_period=3, base_times=(6, 18)),
        ("rr", "oper"): dict(base_times=(0, 3, 6, 9, 12, 15, 18, 21)),
        ("l5", "oper"): dict(data_accumulation_period=1, base_times=(0,)),
    }

    kwargs = KWARGS.get((class_, stream), {})

    context.trace("🌧️", f"accumulations {request} {user_accumulation_period} {kwargs}")

    return _compute_accumulations(
        context,
        dates,
        request,
        user_accumulation_period=user_accumulation_period,
        use_cdsapi_dataset=use_cdsapi_dataset,
        **kwargs,
    )


execute = accumulations

if __name__ == "__main__":
    import yaml

    config = yaml.safe_load(
        """
      class: ea
      expver: '0001'
      grid: 20./20.
      levtype: sfc
#      number: [0, 1]
#      stream: enda
      param: [cp, tp]
#      accumulation_period: 6h
    """
    )
    dates = yaml.safe_load("[2022-12-30 18:00, 2022-12-31 00:00, 2022-12-31 06:00, 2022-12-31 12:00]")
    dates = to_datetime_list(dates)

    for f in accumulations(None, dates, **config):
        print(f, f.to_numpy().mean())



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Any

from . import Command

LOG = logging.getLogger(__name__)


class Publish(Command):
    """Publish a dataset."""

    # This is a command that is used to publish a dataset.
    # it is a class, inheriting from Command.

    internal = True
    timestamp = True

    def add_arguments(self, parser: Any) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        parser : Any
            The command parser to which arguments are added.
        """
        parser.add_argument("path", help="Path of the dataset to publish.")

    def run(self, args: Any) -> None:
        """Execute the publish command.

        Parameters
        ----------
        args : Any
            The arguments passed to the command.
        """
        try:
            from anemoi.registry import publish_dataset
        except ImportError:
            LOG.error("anemoi-registry is not installed. Please install it to use this command.")
            return

        publish_dataset(args.path)


command = Publish



#!/usr/bin/env python3
# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import argparse
import os
import shutil

import xarray as xr

parser = argparse.ArgumentParser(description="Create a sample dataset")
parser.add_argument("input", type=str, help="Input file name")
parser.add_argument("output", type=str, help="Output file name")
args = parser.parse_args()

if os.path.exists(args.output):
    if os.path.isdir(args.output):
        shutil.rmtree(args.output)
    else:
        os.unlink(args.output)

if args.input.endswith(".zarr"):
    ds = xr.open_zarr(args.input)
else:
    ds = xr.open_dataset(args.input)

if args.output.endswith(".zarr"):
    ds.to_zarr(args.output, consolidated=True)
else:
    ds.to_netcdf(args.output)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import re
from abc import ABC
from abc import abstractmethod
from functools import wraps
from typing import Any
from typing import Callable
from typing import List

from .context import Context

LOG = logging.getLogger(__name__)


def notify_result(method: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to notify the context of the result of the method call.

    Parameters
    ----------
    method : Callable[..., Any]
        The method to wrap.

    Returns
    -------
    Callable[..., Any]
        The wrapped method.
    """

    @wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        result: Any = method(self, *args, **kwargs)
        self.context.notify_result(self.action_path, result)
        return result

    return wrapper


class Substitution(ABC):
    """Abstract base class for substitutions in templates."""

    @abstractmethod
    def resolve(self, context: Context) -> Any:
        """Resolve the substitution using the given context.

        Parameters
        ----------
        context : Context
            The context to use for resolution.

        Returns
        -------
        Any
            The resolved value.
        """
        pass


class Reference(Substitution):
    """A class to represent a reference to another value in the context."""

    def __init__(self, context: Any, action_path: List[str]) -> None:
        """Initialize a Reference instance.

        Parameters
        ----------
        context : Any
            The context in which the reference exists.
        action_path : list of str
            The action path to resolve.
        """
        self.context: Any = context
        self.action_path: List[str] = action_path

    def resolve(self, context: Context) -> Any:
        """Resolve the reference using the given context.

        Parameters
        ----------
        context : Context
            The context to use for resolution.

        Returns
        -------
        Any
            The resolved value.
        """
        return context.get_result(self.action_path)


def resolve(context: Context, x: Any) -> Any:
    """Recursively resolve substitutions in the given structure using the context.

    Parameters
    ----------
    context : Context
        The context to use for resolution.
    x : Union[tuple, list, dict, Substitution, Any]
        The structure to resolve.

    Returns
    -------
    Any
        The resolved structure.
    """
    if isinstance(x, tuple):
        return tuple([resolve(context, y) for y in x])

    if isinstance(x, list):
        return [resolve(context, y) for y in x]

    if isinstance(x, dict):
        return {k: resolve(context, v) for k, v in x.items()}

    if isinstance(x, Substitution):
        return x.resolve(context)

    return x


def substitute(context: Context, x: Any) -> Any:
    """Recursively substitute references in the given structure using the context.

    Parameters
    ----------
    context : Context
        The context to use for substitution.
    x : Union[tuple, list, dict, str, Any]
        The structure to substitute.

    Returns
    -------
    Any
        The substituted structure.
    """
    if isinstance(x, tuple):
        return tuple([substitute(context, y) for y in x])

    if isinstance(x, list):
        return [substitute(context, y) for y in x]

    if isinstance(x, dict):
        return {k: substitute(context, v) for k, v in x.items()}

    if not isinstance(x, str):
        return x

    if re.match(r"^\${[\.\w\-]+}$", x):
        path = x[2:-1].split(".")
        context.will_need_reference(path)
        return Reference(context, path)

    return x



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import logging
from typing import Any

from .action import Action
from .action import action_factory
from .step import step_factory
from .trace import trace_select

LOG = logging.getLogger(__name__)


class PipeAction(Action):
    """A class to represent a pipeline of actions."""

    def __init__(self, context: Any, action_path: list, *configs: dict) -> None:
        """Initialize the PipeAction.

        Parameters
        ----------
        context : Any
            The context for the action.
        action_path : list
            The path of the action.
        configs : dict
            The configurations for the actions.
        """
        super().__init__(context, action_path, *configs)
        if len(configs) <= 1:
            raise ValueError(
                f"PipeAction requires at least two actions, got {len(configs)}\n{json.dumps(configs, indent=2)}"
            )

        current: Any = action_factory(configs[0], context, action_path + ["0"])
        for i, c in enumerate(configs[1:]):
            current = step_factory(c, context, action_path + [str(i + 1)], previous_step=current)
        self.last_step: Any = current

    @trace_select
    def select(self, group_of_dates: Any) -> Any:
        """Select data based on the group of dates.

        Parameters
        ----------
        group_of_dates : Any
            The group of dates to select data for.

        Returns
        -------
        Any
            The selected data.
        """
        return self.last_step.select(group_of_dates)

    def __repr__(self) -> str:
        """Return a string representation of the PipeAction."""
        return f"PipeAction({self.last_step})"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any
from typing import List

import earthkit.data as ekd

from .legacy import legacy_source


@legacy_source(__file__)
def execute(context: Any, dates: List[str], **kwargs: Any) -> ekd.FieldList:
    """Executes the loading of an empty data source.

    Parameters
    ----------
    context : object
        The context in which the function is executed.
    dates : list
        List of dates for which data is to be loaded.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    ekd.FieldList
        Loaded empty data source.
    """
    return ekd.from_source("empty")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import datetime
import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import xarray as xr
from earthkit.data.utils.dates import to_datetime

LOG = logging.getLogger(__name__)


def is_scalar(variable: Any) -> bool:
    """Check if the variable is scalar.

    Parameters
    ----------
    variable : Any
        The variable to check.

    Returns
    -------
    bool
        True if the variable is scalar, False otherwise.
    """
    shape = variable.shape
    if shape == (1,):
        return True
    if len(shape) == 0:
        return True
    return False


def extract_single_value(variable: Any) -> Any:
    """Extract a single value from the variable.

    Parameters
    ----------
    variable : Any
        The variable to extract the value from.

    Returns
    -------
    Any
        The extracted single value.
    """
    shape = variable.shape
    if np.issubdtype(variable.values.dtype, np.datetime64):
        if len(shape) == 0:
            return to_datetime(variable.values)  # Convert to python datetime
        if shape == (1,):
            return to_datetime(variable.values[0])
        assert False, (shape, variable.values[:2])

    if np.issubdtype(variable.values.dtype, np.timedelta64):
        if len(shape) == 0:
            # Convert to python timedelta64
            return datetime.timedelta(seconds=variable.values.astype("timedelta64[s]").astype(int).item())
        assert False, (shape, variable.values)

    if shape == (1,):
        return variable.values[0]

    if len(shape) == 0:
        return variable.values.item()

    assert False, (shape, variable.values)


class Coordinate:
    """Base class for coordinates."""

    is_grid = False
    is_dim = True
    is_lat = False
    is_lon = False
    is_time = False
    is_step = False
    is_date = False
    is_member = False
    is_x = False
    is_y = False

    def __init__(self, variable: xr.DataArray) -> None:
        """Initialize the coordinate.

        Parameters
        ----------
        variable : Any
            The variable representing the coordinate.
        """
        self.variable = variable
        self.scalar = is_scalar(variable)
        self.kwargs: Dict[str, Any] = {}  # Used when creating a new coordinate (reduced method)

    def __len__(self) -> int:
        """Get the length of the coordinate.

        Returns
        -------
        int
            The length of the coordinate.
        """
        return 1 if self.scalar else len(self.variable)

    def __repr__(self) -> str:
        """Get the string representation of the coordinate.

        Returns
        -------
        str
            The string representation of the coordinate.
        """
        return "%s[name=%s,values=%s,shape=%s]" % (
            self.__class__.__name__,
            self.variable.name,
            self.variable.values if self.scalar else len(self),
            self.variable.shape,
        )

    def reduced(self, i: int) -> Coordinate:
        """Create a new coordinate with a single value.

        Parameters
        ----------
        i : int
            The index of the value to select.

        Returns
        -------
        Coordinate
            A new coordinate with the selected value.
        """
        return self.__class__(
            self.variable.isel({self.variable.dims[0]: i}),
            **self.kwargs,
        )

    def index(self, value: Union[Any, list, tuple]) -> Optional[Union[int, list]]:
        """Return the index of the value in the coordinate.

        Parameters
        ----------
        value : Union[Any, list, tuple]
            The value to search for.

        Returns
        -------
        Optional[Union[int, list]]
            The index or indices of the value in the coordinate, or None if not found.
        """
        if isinstance(value, (list, tuple)):
            if len(value) == 1:
                return self._index_single(value)
            else:
                return self._index_multiple(value)
        return self._index_single(value)

    def _index_single(self, value: Any) -> Optional[int]:
        """Return the index of a single value in the coordinate.

        Parameters
        ----------
        value : Any
            The value to search for.

        Returns
        -------
        Optional[int]
            The index of the value in the coordinate, or None if not found.
        """
        values = self.variable.values

        # Check if dimension is 0D
        if not isinstance(values, (list, np.ndarray)):
            values = [values]

        # Assume the array is sorted
        index = np.searchsorted(values, value)

        if index < len(values) and values[index] == value:
            return index

        # If not found, we need to check if the value is in the array

        index = np.where(values == value)[0]
        if len(index) > 0:
            return index[0]

        return None

    def _index_multiple(self, value: list) -> Optional[list]:
        """Return the indices of multiple values in the coordinate.

        Parameters
        ----------
        value : list
            The values to search for.

        Returns
        -------
        Optional[list]
            The indices of the values in the coordinate, or None if not found.
        """
        values = self.variable.values

        # Check if dimension is 0D
        if not isinstance(values, (list, np.ndarray)):
            values = [values]

        # Assume the array is sorted

        index = np.searchsorted(values, value)
        index = index[index < len(values)]

        if np.all(values[index] == value):
            return index

        # If not found, we need to check if the value is in the array

        index = np.where(np.isin(values, value))[0]

        # We could also return incomplete matches
        if len(index) == len(value):
            return index

        return None

    @property
    def name(self) -> str:
        """Get the name of the coordinate."""
        return self.variable.name

    def normalise(self, value: Any) -> Any:
        """Normalize the value for the coordinate.

        Parameters
        ----------
        value : Any
            The value to normalize.

        Returns
        -------
        Any
            The normalized value.
        """
        # Subclasses to format values that will be added to the field metadata
        return value

    @property
    def single_value(self) -> Any:
        """Get the single value of the coordinate."""
        return extract_single_value(self.variable)


class TimeCoordinate(Coordinate):
    """Coordinate class for time."""

    is_time = True
    mars_names = ("valid_datetime",)

    def index(self, time: datetime.datetime) -> Optional[int]:
        """Return the index of the time in the coordinate.

        Parameters
        ----------
        time : datetime.datetime
            The time to search for.

        Returns
        -------
        Optional[int]
            The index of the time in the coordinate, or None if not found.
        """
        return super().index(np.datetime64(time))


class DateCoordinate(Coordinate):
    """Coordinate class for date."""

    is_date = True
    mars_names = ("date",)

    def index(self, date: datetime.datetime) -> Optional[int]:
        """Return the index of the date in the coordinate.

        Parameters
        ----------
        date : datetime.datetime
            The date to search for.

        Returns
        -------
        Optional[int]
            The index of the date in the coordinate, or None if not found.
        """
        return super().index(np.datetime64(date))


class StepCoordinate(Coordinate):
    """Coordinate class for step."""

    is_step = True
    mars_names = ("step",)


class LevelCoordinate(Coordinate):
    """Coordinate class for level.

    Parameters
    ----------
    variable : Any
        The variable representing the coordinate.
    levtype : str
        The type of level.
    """

    mars_names = ("level", "levelist")

    def __init__(self, variable: Any, levtype: str) -> None:
        """Initialize the level coordinate.

        Parameters
        ----------
        variable : Any
            The variable representing the coordinate.
        levtype : str
            The type of level.
        """
        super().__init__(variable)
        self.levtype = levtype
        # kwargs is used when creating a new coordinate (reduced method)
        self.kwargs = {"levtype": levtype}

    def normalise(self, value: Any) -> Any:
        """Normalize the value for the level coordinate.

        Parameters
        ----------
        value : Any
            The value to normalize.

        Returns
        -------
        Any
            The normalized value.
        """
        # Some netcdf have pressue levels in float
        if int(value) == value:
            return int(value)
        return value


class EnsembleCoordinate(Coordinate):
    """Coordinate class for ensemble."""

    is_member = True
    mars_names = ("number",)

    def normalise(self, value: Any) -> Any:
        """Normalize the value for the ensemble coordinate.

        Parameters
        ----------
        value : Any
            The value to normalize.

        Returns
        -------
        Any
            The normalized value.
        """
        if int(value) == value:
            return int(value)
        return value


class LongitudeCoordinate(Coordinate):
    """Coordinate class for longitude."""

    is_grid = True
    is_lon = True
    mars_names = ("longitude",)


class LatitudeCoordinate(Coordinate):
    """Coordinate class for latitude."""

    is_grid = True
    is_lat = True
    mars_names = ("latitude",)


class XCoordinate(Coordinate):
    """Coordinate class for X."""

    is_grid = True
    is_x = True
    mars_names = ("x",)


class YCoordinate(Coordinate):
    """Coordinate class for Y."""

    is_grid = True
    is_y = True
    mars_names = ("y",)


class ScalarCoordinate(Coordinate):
    """Coordinate class for scalar."""

    is_grid = False

    @property
    def mars_names(self) -> Tuple[str, ...]:
        """Get the MARS names for the coordinate."""
        return (self.variable.name,)


class UnsupportedCoordinate(Coordinate):
    """Coordinate class for unsupported coordinates."""

    @property
    def mars_names(self) -> tuple:
        """Get the MARS names for the coordinate."""
        return (self.variable.name,)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import xarray as xr
import yaml
from earthkit.data import FieldList

from .field import EmptyFieldList
from .flavour import CoordinateGuesser
from .patch import patch_dataset
from .time import Time
from .variable import FilteredVariable
from .variable import Variable

LOG = logging.getLogger(__name__)


class XarrayFieldList(FieldList):
    """A class to represent a list of fields from an xarray Dataset."""

    def __init__(self, ds: xr.Dataset, variables: List[Variable]) -> None:
        """Initialize the XarrayFieldList.

        Parameters
        ----------
        ds : xr.Dataset
            The xarray Dataset.
        variables : List[Variable]
            The list of variables.
        """
        self.ds: xr.Dataset = ds
        self.variables: List[Variable] = variables.copy()
        self.total_length: int = sum(v.length for v in variables)

    def __repr__(self) -> str:
        """Return a string representation of the XarrayFieldList."""
        return f"XarrayFieldList({self.total_length})"

    def __len__(self) -> int:
        """Return the length of the XarrayFieldList."""
        return self.total_length

    def __getitem__(self, i: int) -> Any:
        """Get an item from the XarrayFieldList by index.

        Parameters
        ----------
        i : int
            The index of the item to get.

        Returns
        -------
        Any
            The item at the specified index.

        Raises
        ------
        IndexError
            If the index is out of range.
        """
        k: int = i

        if i < 0:
            i = self.total_length + i

        for v in self.variables:
            if i < v.length:
                return v[i]
            i -= v.length

        raise IndexError(k)

    @classmethod
    def from_xarray(
        cls,
        ds: xr.Dataset,
        *,
        flavour: Optional[Union[str, Dict[str, Any]]] = None,
        patch: Optional[Dict[str, Any]] = None,
    ) -> "XarrayFieldList":
        """Create an XarrayFieldList from an xarray Dataset.

        Parameters
        ----------
        ds : xr.Dataset
            The xarray Dataset to create the field list from.
        flavour : Optional[Union[str, Dict[str, Any]]], optional
            The flavour to use for guessing coordinates.
        patch : Optional[Dict[str, Any]], optional
            The patch to apply to the dataset.

        Returns
        -------
        XarrayFieldList
            The created XarrayFieldList.
        """
        if patch is not None:
            ds = patch_dataset(ds, patch)

        variables: List[Variable] = []

        if isinstance(flavour, str):
            with open(flavour) as f:
                if flavour.endswith(".yaml") or flavour.endswith(".yml"):
                    flavour = yaml.safe_load(f)
                else:
                    flavour = json.load(f)

        if isinstance(flavour, Dict):
            flavour_coords: List[str] = [coords["name"] for coords in flavour["rules"].values()]
            ds_dims: List[str] = [dim for dim in ds._dims]
            for dim in ds_dims:
                if dim in flavour_coords and dim not in ds._coord_names:
                    ds = ds.assign_coords({dim: ds[dim]})
                else:
                    pass

        guess = CoordinateGuesser.from_flavour(ds, flavour)

        skip: set = set()

        def _skip_attr(v: Any, attr_name: str) -> None:
            attr_val: str = getattr(v, attr_name, "")
            if isinstance(attr_val, str):
                skip.update(attr_val.split(" "))

        for name in ds.data_vars:
            variable = ds[name]
            _skip_attr(variable, "coordinates")
            _skip_attr(variable, "bounds")
            _skip_attr(variable, "grid_mapping")

        LOG.debug("Xarray data_vars: %s", ds.data_vars)

        # Select only geographical variables
        for name in ds.data_vars:

            if name in skip:
                continue

            variable = ds[name]
            coordinates: List[Any] = []

            for coord in variable.coords:

                c = guess.guess(ds[coord], coord)
                assert c, f"Could not guess coordinate for {coord}"
                if coord not in variable.dims:
                    LOG.debug("%s: coord=%s (not a dimension): dims=%s", variable, coord, variable.dims)
                    c.is_dim = False
                coordinates.append(c)

            grid_coords: int = sum(1 for c in coordinates if c.is_grid)
            # assert grid_coords <= 2, [c for c in coordinates if c.is_grid]

            if grid_coords < 2:
                LOG.debug("Skipping %s (not 2D): %s", variable, [(c, c.is_grid, c.is_dim) for c in coordinates])
                continue

            v = Variable(
                ds=ds,
                variable=variable,
                coordinates=coordinates,
                grid=guess.grid(coordinates, variable),
                time=Time.from_coordinates(coordinates),
                metadata={},
            )

            variables.append(v)

        return cls(ds, variables)

    def sel(self, **kwargs: Any) -> FieldList:
        """Select fields from the XarrayFieldList based on criteria.

        Parameters
        ----------
        kwargs : dict
            The selection criteria.

        Returns
        -------
        FieldList
            The new FieldList with selected fields.
        """

        """
        The algorithm is as follows:
        1 - Use the kwargs to select the variables that match the selection (`param` or `variable`)
        2 - For each variable, use the remaining kwargs to select the coordinates (`level`, `number`, ...)
        3 - Some mars like keys, like `date`, `time`, `step` are not found in the coordinates,
            but added to the metadata of the selected fields. A example is `step` that is added to the
            metadata of the field. Step 2 may return a variable that contain all the fields that
            verify at the same `valid_datetime`, with different base `date` and `time` and a different `step`.
            So we get an extra chance to filter the fields by the metadata.
        """

        variables: List[Variable] = []
        count: int = 0

        for v in self.variables:

            # First, select matching variables

            # This will consume 'param' or 'variable' from kwargs
            # and return the rest
            match, rest = v.match(**kwargs)

            if match:
                count += 1
                missing: Dict[str, Any] = {}

                # Select from the variable's coordinates (time, level, number, ....)
                # This may return a new variable with a isel() slice of the selection
                # or None if the selection is not possible. In this case, missing is updated
                # with the values of kwargs (rest) that are not relevant for this variable
                v = v.sel(missing, **rest)
                if missing:
                    if v is not None:
                        # The remaining kwargs are passed used to create a FilteredVariable
                        # that will select 2D slices based on their metadata
                        v = FilteredVariable(v, **missing)
                    else:
                        LOG.warning(f"Variable {v} has missing coordinates: {missing}")

                if v is not None:
                    variables.append(v)

        if count == 0:
            LOG.warning("No variable found for %s", kwargs)
            LOG.warning("Variables: %s", sorted([v.name for v in self.variables]))

        if not variables:
            return EmptyFieldList()

        return self.__class__(self.ds, variables)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import sqlite3
from typing import Any
from typing import Iterator
from typing import List
from typing import Optional

import earthkit.data as ekd
import tqdm
from anemoi.transform.flavour import RuleBasedFlavour
from cachetools import LRUCache
from earthkit.data.indexing.fieldlist import FieldArray

from .legacy import legacy_source

LOG = logging.getLogger(__name__)

KEYS1 = ("class", "type", "stream", "expver", "levtype")
KEYS2 = ("shortName", "paramId", "level", "step", "number", "date", "time", "valid_datetime", "levelist")

KEYS = KEYS1 + KEYS2


class GribIndex:
    def __init__(
        self,
        database: str,
        *,
        keys: Optional[List[str] | str] = None,
        flavour: Optional[str] = None,
        update: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Initialize the GribIndex object.

        Parameters
        ----------
        database : str
            Path to the SQLite database file.
        keys : Optional[List[str] | str], optional
            List of keys or a string of keys to use for indexing, by default None.
        flavour : Optional[str], optional
            Flavour configuration for mapping fields, by default None.
        update : bool, optional
            Whether to update the database, by default False.
        overwrite : bool, optional
            Whether to overwrite the database if it exists, by default False.
        """
        self.database = database
        if overwrite:
            assert update
            if os.path.exists(database):
                os.remove(database)

        if not update:
            if not os.path.exists(database):
                raise FileNotFoundError(f"Database {database} does not exist")

        if keys is not None:
            if isinstance(keys, str):
                if keys.startswith("+"):
                    keys = set(KEYS) | set(keys[1:].split(","))
                else:
                    keys = set(",".split(keys.split(",")))
                keys = list(keys)

        self.conn = sqlite3.connect(database)
        self.cursor = self.conn.cursor()

        if flavour is not None:
            self.flavour = RuleBasedFlavour(flavour)
        else:
            self.flavour = None

        self.update = update
        self.cache = None
        self.keys = keys
        self._columns = None

        if update:
            if self.keys is None:
                self.keys = KEYS
            LOG.info(f"Using keys: {sorted(self.keys)}")
            self._create_tables()
        else:
            assert keys is None
            self.keys = self._all_columns()
            self.cache = LRUCache(maxsize=50)

        self.warnings = {}
        self.cache = {}

    def _create_tables(self) -> None:
        """Create the necessary tables in the database."""
        assert self.update

        self.cursor.execute(
            """
        CREATE TABLE IF NOT EXISTS paths (
            id INTEGER PRIMARY KEY,
            path TEXT not null
        )
        """
        )

        columns = ("valid_datetime",)
        # We don't use NULL as a default because NULL is considered a different value
        # in UNIQUE INDEX constraints (https://www.sqlite.org/lang_createindex.html)

        self.cursor.execute(
            f"""
        CREATE TABLE IF NOT EXISTS grib_index (
            _id INTEGER PRIMARY KEY,
            _path_id INTEGER not null,
            _offset INTEGER not null,
            _length INTEGER not null,
            {', '.join(f"{key} TEXT not null default ''" for key in columns)},
            FOREIGN KEY(_path_id) REFERENCES paths(id))
        """
        )  # ,

        self.cursor.execute(
            """
        CREATE UNIQUE INDEX IF NOT EXISTS idx_grib_index_path_offset
        ON grib_index (_path_id, _offset)
        """
        )

        self.cursor.execute(
            f"""
        CREATE UNIQUE INDEX IF NOT EXISTS idx_grib_index_all_keys
        ON grib_index ({', '.join(columns)})
        """
        )

        for key in columns:
            self.cursor.execute(
                f"""
            CREATE INDEX IF NOT EXISTS idx_grib_index_{key}
            ON grib_index ({key})
            """
            )

        self._commit()

    def _commit(self) -> None:
        """Commit the current transaction to the database."""
        self.conn.commit()

    def _get_metadata_keys(self) -> List[str]:
        """Retrieve the metadata keys from the database.

        Returns
        -------
        List[str]
            A list of metadata keys stored in the database.
        """
        self.cursor.execute("SELECT key FROM metadata_keys")
        return [row[0] for row in self.cursor.fetchall()]

    def _path_id(self, path: str) -> int:
        """Get the id of a path in the database.

        Parameters
        ----------
        path : str
            The file path to retrieve or insert.

        Returns
        -------
        int
            The ID of the path in the database.
        """
        self.cursor.execute("SELECT id FROM paths WHERE path = ?", (path,))
        row = self.cursor.fetchone()
        if row is None:
            self.cursor.execute("INSERT INTO paths (path) VALUES (?)", (path,))
            self._commit()
            return self.cursor.lastrowid
        return row[0]

    def _add_grib(self, **kwargs: Any) -> None:
        """Add a GRIB record to the database.

        Parameters
        ----------
        **kwargs : Any
            Key-value pairs representing the GRIB record fields.
        """
        assert self.update

        try:

            self.cursor.execute(
                f"""
            INSERT INTO grib_index ({', '.join(kwargs.keys())})
            VALUES ({', '.join('?' for _ in kwargs)})
            """,
                tuple(kwargs.values()),
            )

        except sqlite3.IntegrityError:
            LOG.error(f"Error adding grib record: {kwargs}")
            LOG.error("Record already exists")
            LOG.info(f"Path: {self._get_path(kwargs['_path_id'])}")
            for n in ("_path_id", "_offset", "_length"):
                kwargs.pop(n)
            self.cursor.execute(
                "SELECT * FROM grib_index WHERE " + " AND ".join(f"{key} = ?" for key in kwargs.keys()),
                tuple(kwargs.values()),
            )
            existing_record = self.cursor.fetchone()
            if existing_record:
                LOG.info(f"Existing record found: {existing_record}")
                LOG.info(f"Path: {self._get_path(existing_record[1])}")
            raise

    def _all_columns(self) -> List[str]:
        """Retrieve all column names from the grib_index table.

        Returns
        -------
        List[str]
            A list of column names.
        """
        if self._columns is not None:
            return self._columns

        self.cursor.execute("PRAGMA table_info(grib_index)")
        columns = {row[1] for row in self.cursor.fetchall()}
        self._columns = [col for col in columns if not col.startswith("_")]
        return self._columns

    def _ensure_columns(self, columns: List[str]) -> None:
        """Add missing columns to the grib_index table.

        Parameters
        ----------
        columns : List[str]
            List of column names to ensure in the table.
        """
        assert self.update

        existing_columns = self._all_columns()
        new_columns = [column for column in columns if column not in existing_columns]

        if not new_columns:
            return

        self._columns = None

        for column in new_columns:
            self.cursor.execute(f"ALTER TABLE grib_index ADD COLUMN {column} TEXT not null default ''")

        self.cursor.execute("""DROP INDEX IF EXISTS idx_grib_index_all_keys""")
        all_columns = self._all_columns()

        self.cursor.execute(
            f"""
        CREATE UNIQUE INDEX IF NOT EXISTS idx_grib_index_all_keys
        ON grib_index ({', '.join(all_columns)})
        """
        )

        for key in all_columns:
            self.cursor.execute(
                f"""
            CREATE INDEX IF NOT EXISTS idx_grib_index_{key}
            ON grib_index ({key})
            """
            )

    def add_grib_file(self, path: str) -> None:
        """Add a GRIB file to the database.

        Parameters
        ----------
        path : str
            Path to the GRIB file to add.
        """
        path_id = self._path_id(path)

        fields = ekd.from_source("file", path)
        if self.flavour is not None:
            fields = self.flavour.map(fields)

        for i, field in enumerate(tqdm.tqdm(fields, leave=False)):

            keys = field.metadata(namespace="mars").copy()
            keys.update({k: field.metadata(k, default=None) for k in self.keys})

            keys.setdefault("param", keys.get("shortName", keys.get("paramId")))

            keys = {k: v for k, v in keys.items() if v is not None}

            if keys.get("param") in (0, "unknown"):
                param = (
                    field.metadata("discipline", default=None),
                    field.metadata("parameterCategory", default=None),
                    field.metadata("parameterNumber", default=None),
                )
                if param not in self.warnings:
                    self._unknown(path, field, i, param)
                    self.warnings[param] = True

            self._ensure_columns(list(keys.keys()))

            self._add_grib(
                _path_id=path_id,
                _offset=field.metadata("offset"),
                _length=field.metadata("totalLength"),
                **keys,
            )

        self._commit()

    def _paramdb(self, category: int, discipline: int) -> Optional[dict]:
        """Fetch parameter information from the parameter database.

        Parameters
        ----------
        category : int
            The parameter category.
        discipline : int
            The parameter discipline.

        Returns
        -------
        Optional[dict]
            The parameter information, or None if unavailable.
        """
        if (category, discipline) in self.cache:
            return self.cache[(category, discipline)]

        try:
            import requests

            r = requests.get(
                f"https://codes.ecmwf.int/parameter-database/api/v1/param?category={category}&discipline={discipline}"
            )
            r.raise_for_status()
            self.cache[(category, discipline)] = r.json()
            return self.cache[(category, discipline)]

        except Exception as e:
            LOG.warning(f"Failed to fetch information from parameter database: {e}")

    def _param_grib2_info(self, paramId: int) -> List[dict]:
        """Fetch GRIB2 parameter information for a given parameter ID.

        Parameters
        ----------
        paramId : int
            The parameter ID.

        Returns
        -------
        List[dict]
            A list of GRIB2 parameter information.
        """
        if ("grib2", paramId) in self.cache:
            return self.cache[("grib2", paramId)]

        try:
            import requests

            r = requests.get(f"https://codes.ecmwf.int/parameter-database/api/v1/param/{paramId}/grib2/")
            r.raise_for_status()
            self.cache[("grib2", paramId)] = r.json()
            return self.cache[("grib2", paramId)]

        except Exception as e:
            LOG.warning(f"Failed to fetch information from parameter database: {e}")
        return []

    def _param_id_info(self, paramId: int) -> Optional[dict]:
        """Fetch detailed information for a given parameter ID.

        Parameters
        ----------
        paramId : int
            The parameter ID.

        Returns
        -------
        Optional[dict]
            The parameter information, or None if unavailable.
        """
        if ("info", paramId) in self.cache:
            return self.cache[("info", paramId)]

        try:
            import requests

            r = requests.get(f"https://codes.ecmwf.int/parameter-database/api/v1/param/{paramId}/")
            r.raise_for_status()
            self.cache[("info", paramId)] = r.json()
            return self.cache[("info", paramId)]

        except Exception as e:
            LOG.warning(f"Failed to fetch information from parameter database: {e}")

        return None

    def _param_id_unit(self, unitId: int) -> Optional[dict]:
        """Fetch unit information for a given unit ID.

        Parameters
        ----------
        unitId : int
            The unit ID.

        Returns
        -------
        Optional[dict]
            The unit information, or None if unavailable.
        """
        if ("unit", unitId) in self.cache:
            return self.cache[("unit", unitId)]

        try:
            import requests

            r = requests.get(f"https://codes.ecmwf.int/parameter-database/api/v1/unit/{unitId}/")
            r.raise_for_status()
            self.cache[("unit", unitId)] = r.json()
            return self.cache[("unit", unitId)]

        except Exception as e:
            LOG.warning(f"Failed to fetch information from parameter database: {e}")

        return None

    def _unknown(self, path: str, field: ekd.Field, i: int, param: tuple) -> None:
        """Log information about unknown parameters.

        Parameters
        ----------
        path : str
            Path to the GRIB file.
        field : ekd.Field
            The GRIB field object.
        i : int
            The index of the field in the file.
        param : tuple
            The parameter tuple (discipline, category, parameterNumber).
        """

        def _(s):
            try:
                return int(s)
            except ValueError:
                return s

        LOG.warning(
            f"Unknown param for message {i+1} in {path} at offset {int(field.metadata('offset', default=None))}"
        )
        LOG.warning(
            f"shortName/paramId: {field.metadata('shortName', default=None)}/{field.metadata('paramId', default=None)}"
        )
        name = field.metadata("parameterName", default=None)
        units = field.metadata("parameterUnits", default=None)
        LOG.warning(f"Discipline/category/parameter: {param} ({name}, {units})")
        LOG.warning(f"grib_copy -w count={i+1} {path} tmp.grib")

        info = self._paramdb(discipline=param[0], category=param[1])
        found = set()
        if info is not None:
            for n in tqdm.tqdm(info, desc="Scanning parameter database"):

                for p in self._param_grib2_info(n["id"]):

                    keys = {k["name"]: _(k["value"]) for k in p["keys"]}
                    if keys.get("parameterNumber") == param[2]:
                        found.add(n["id"])

        for n in found:
            info = self._param_id_info(n)
            if "unit_id" in info:
                info["unit_id"] = self._param_id_unit(info["unit_id"])["name"]

            LOG.info("%s", f"Possible match: {n}")
            LOG.info("%s", f"     Name:        {info.get('name')}")
            LOG.info("%s", f"     Short name:  {info.get('shortname')}")
            LOG.info("%s", f"     Units:       {info.get('unit_id')}")
            LOG.info("%s", f"     Description: {info.get('description')}")
            LOG.info("")

    def _get_path(self, path_id: int) -> str:
        """Retrieve the path corresponding to a given path_id.

        Parameters
        ----------
        path_id : int
            The ID of the path to retrieve.

        Returns
        -------
        str
            The path corresponding to the given path_id.

        Raises
        ------
        ValueError
            If the path_id does not exist in the database.
        """
        self.cursor.execute("SELECT path FROM paths WHERE id = ?", (path_id,))
        row = self.cursor.fetchone()
        if row is None:
            raise ValueError(f"No path found for path_id {path_id}")
        return row[0]

    def retrieve(self, dates: List[Any], **kwargs: Any) -> Iterator[Any]:
        """Retrieve GRIB data from the database.

        Parameters
        ----------
        dates : List[Any]
            List of dates to retrieve data for.
        **kwargs : Any
            Additional filtering criteria.

        Returns
        ------
        Iterator[Any]
            The GRIB data matching the criteria.
        """
        assert not self.update

        dates = [d.isoformat() for d in dates]

        query = """SELECT _path_id, _offset, _length
                   FROM grib_index WHERE valid_datetime IN ({})""".format(
            ", ".join("?" for _ in dates)
        )
        params = dates

        for k, v in kwargs.items():
            if isinstance(v, list):
                query += f" AND {k} IN ({', '.join('?' for _ in v)})"
                params.extend([str(_) for _ in v])
            else:
                query += f" AND {k} = ?"
                params.append(str(v))

        print("SELECT", query)
        print("SELECT", params)

        self.cursor.execute(query, params)
        for path_id, offset, length in self.cursor.fetchall():
            if path_id in self.cache:
                file = self.cache[path_id]
            else:
                path = self._get_path(path_id)
                LOG.info(f"Opening {path}")
                self.cache[path_id] = open(path, "rb")
                file = self.cache[path_id]

            file.seek(offset)
            data = file.read(length)
            yield data


@legacy_source(__file__)
def execute(
    context: Any,
    dates: List[Any],
    indexdb: str,
    flavour: Optional[str] = None,
    **kwargs: Any,
) -> FieldArray:
    """Execute the GRIB data retrieval process.

    Parameters
    ----------
    context : Any
        The execution context.
    dates : List[Any]
        List of dates to retrieve data for.
    indexdb : str
        Path to the GRIB index database.
    flavour : Optional[str], optional
        Flavour configuration for mapping fields, by default None.
    **kwargs : Any
        Additional filtering criteria.

    Returns
    -------
    FieldArray
        An array of retrieved GRIB fields.
    """
    index = GribIndex(indexdb)
    result = []

    if flavour is not None:
        flavour = RuleBasedFlavour(flavour)

    for grib in index.retrieve(dates, **kwargs):
        field = ekd.from_source("memory", grib)[0]
        if flavour:
            field = flavour.apply(field)
        result.append(field)

    return FieldArray(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod
from functools import cached_property
from typing import Any
from typing import Tuple

import numpy as np

LOG = logging.getLogger(__name__)


class Grid(ABC):
    """Abstract base class for grid structures."""

    def __init__(self) -> None:
        pass

    @property
    def latitudes(self) -> Any:
        """Get the latitudes of the grid."""
        return self.grid_points[0]

    @property
    def longitudes(self) -> Any:
        """Get the longitudes of the grid."""
        return self.grid_points[1]

    @property
    @abstractmethod
    def grid_points(self) -> Tuple[Any, Any]:
        """Get the grid points."""
        pass


class LatLonGrid(Grid):
    """Grid class for latitude and longitude coordinates."""

    def __init__(self, lat: Any, lon: Any, variable_dims: Any) -> None:
        """Initialize the LatLonGrid class.

        Parameters
        ----------
        lat : Any
            The latitudes.
        lon : Any
            The longitudes.
        variable_dims : Any
            The variable dimensions.
        """
        super().__init__()
        self.lat = lat
        self.lon = lon
        self.variable_dims = variable_dims


class XYGrid(Grid):
    """Grid class for x and y coordinates."""

    def __init__(self, x: Any, y: Any) -> None:
        """Initialize the XYGrid class.

        Parameters
        ----------
        x : Any
            The x-coordinates.
        y : Any
            The y-coordinates.
        """
        self.x = x
        self.y = y


class MeshedGrid(LatLonGrid):
    """Grid class for meshed latitude and longitude coordinates."""

    @cached_property
    def grid_points(self) -> Tuple[Any, Any]:
        """Get the grid points for the meshed grid."""

        if self.variable_dims == (self.lon.variable.name, self.lat.variable.name):
            lat, lon = np.meshgrid(
                self.lat.variable.values,
                self.lon.variable.values,
            )
        elif self.variable_dims == (self.lat.variable.name, self.lon.variable.name):
            lon, lat = np.meshgrid(
                self.lon.variable.values,
                self.lat.variable.values,
            )

        else:
            raise NotImplementedError(f"MeshedGrid.grid_points: unrecognized variable_dims {self.variable_dims}")

        return lat.flatten(), lon.flatten()


class UnstructuredGrid(LatLonGrid):
    """Grid class for unstructured latitude and longitude coordinates."""

    def __init__(self, lat: Any, lon: Any, variable_dims: Any) -> None:
        """Initialize the UnstructuredGrid class.

        Parameters
        ----------
        lat : Any
            The latitudes.
        lon : Any
            The longitudes.
        variable_dims : Any
            The variable dimensions.
        """
        super().__init__(lat, lon, variable_dims)
        assert len(lat) == len(lon), (len(lat), len(lon))
        self.variable_dims = variable_dims
        self.grid_dims = lat.variable.dims
        assert lon.variable.dims == self.grid_dims, (lon.variable.dims, self.grid_dims)
        assert set(self.variable_dims) == set(self.grid_dims), (self.variable_dims, self.grid_dims)

    @cached_property
    def grid_points(self) -> Tuple[Any, Any]:
        """Get the grid points for the unstructured grid."""
        assert 1 <= len(self.variable_dims) <= 2

        if len(self.variable_dims) == 1:
            return self.lat.variable.values.flatten(), self.lon.variable.values.flatten()

        if len(self.variable_dims) == 2 and self.variable_dims == self.grid_dims:
            return self.lat.variable.values.flatten(), self.lon.variable.values.flatten()

        LOG.warning(
            "UnstructuredGrid: variable indexing %s does not match grid indexing %s", self.variable_dims, self.grid_dims
        )

        lat = self.lat.variable.values.transpose().flatten()
        lon = self.lon.variable.values.transpose().flatten()

        return lat, lon


class ProjectionGrid(XYGrid):
    """Grid class for projected x and y coordinates."""

    def __init__(self, x: Any, y: Any, projection: Any) -> None:
        """Initialize the ProjectionGrid class.

        Parameters
        ----------
        x : Any
            The x-coordinates.
        y : Any
            The y-coordinates.
        projection : Any
            The projection information.
        """
        super().__init__(x, y)
        self.projection = projection

    def transformer(self) -> Any:
        """Get the transformer for the projection.

        Returns
        -------
        Any
            The transformer.
        """
        from pyproj import CRS
        from pyproj import Transformer

        if isinstance(self.projection, dict):
            data_crs = CRS.from_cf(self.projection)
        else:
            data_crs = self.projection

        wgs84_crs = CRS.from_epsg(4326)  # WGS84

        return Transformer.from_crs(data_crs, wgs84_crs, always_xy=True)


class MeshProjectionGrid(ProjectionGrid):
    """Grid class for meshed projected coordinates."""

    @cached_property
    def grid_points(self) -> Tuple[Any, Any]:
        """Get the grid points for the mesh projection grid."""
        transformer = self.transformer()
        xv, yv = np.meshgrid(self.x.variable.values, self.y.variable.values)  # , indexing="ij")
        lon, lat = transformer.transform(xv, yv)
        return lat.flatten(), lon.flatten()


class UnstructuredProjectionGrid(XYGrid):
    """Grid class for unstructured projected coordinates."""

    @cached_property
    def grid_points(self) -> Tuple[Any, Any]:
        """Get the grid points for the unstructured projection grid."""
        raise NotImplementedError("UnstructuredProjectionGrid")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import List

import xarray as xr

LOG = logging.getLogger(__name__)


def patch_attributes(ds: xr.Dataset, attributes: Dict[str, Dict[str, Any]]) -> Any:
    """Patch the attributes of the dataset.

    Parameters
    ----------
    ds : xr.Dataset
        The dataset to patch.
    attributes : Dict[str, Dict[str, Any]]
        The attributes to patch.

    Returns
    -------
    Any
        The patched dataset.
    """
    for name, value in attributes.items():
        variable = ds[name]
        variable.attrs.update(value)

    return ds


def patch_coordinates(ds: xr.Dataset, coordinates: List[str]) -> Any:
    """Patch the coordinates of the dataset.

    Parameters
    ----------
    ds : xr.Dataset
        The dataset to patch.
    coordinates : List[str]
        The coordinates to patch.

    Returns
    -------
    Any
        The patched dataset.
    """
    for name in coordinates:
        ds = ds.assign_coords({name: ds[name]})

    return ds


PATCHES = {
    "attributes": patch_attributes,
    "coordinates": patch_coordinates,
}


def patch_dataset(ds: xr.Dataset, patch: Dict[str, Dict[str, Any]]) -> Any:
    """Patch the dataset.

    Parameters
    ----------
    ds : xr.Dataset
        The dataset to patch.
    patch : Dict[str, Dict[str, Any]]
        The patch to apply.

    Returns
    -------
    Any
        The patched dataset.
    """
    for what, values in patch.items():
        if what not in PATCHES:
            raise ValueError(f"Unknown patch type {what!r}")

        ds = PATCHES[what](ds, values)

    return ds



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import xarray as xr
from anemoi.utils.testing import skip_if_offline
from anemoi.utils.testing import skip_missing_packages

from anemoi.datasets.create.sources.xarray import XarrayFieldList
from anemoi.datasets.data.stores import name_to_zarr_store
from anemoi.datasets.testing import assert_field_list


@skip_if_offline
@skip_missing_packages("gcsfs")
def test_arco_era5_1() -> None:
    """Test loading and validating the arco_era5_1 dataset."""
    ds = xr.open_zarr(
        "gs://gcp-public-data-arco-era5/ar/1959-2022-full_37-1h-0p25deg-chunk-1.zarr-v2",
        chunks={"time": 48},
        consolidated=True,
    )

    fs = XarrayFieldList.from_xarray(ds)
    assert_field_list(
        fs,
        128677526,
        "1959-01-01T00:00:00",
        "2021-12-31T23:00:00",
    )


@skip_if_offline
@skip_missing_packages("gcsfs")
def test_arco_era5_2() -> None:
    """Test loading and validating the arco_era5_2 dataset."""
    ds = xr.open_zarr(
        "gs://gcp-public-data-arco-era5/ar/1959-2022-1h-360x181_equiangular_with_poles_conservative.zarr",
        chunks={"time": 48},
        consolidated=True,
    )

    fs = XarrayFieldList.from_xarray(ds)
    assert_field_list(
        fs,
        128677526,
        "1959-01-01T00:00:00",
        "2021-12-31T23:00:00",
    )


@skip_if_offline
@skip_missing_packages("gcsfs")
def test_weatherbench() -> None:
    """Test loading and validating the weatherbench dataset."""
    ds = xr.open_zarr("gs://weatherbench2/datasets/pangu_hres_init/2020_0012_0p25.zarr")

    # https://weatherbench2.readthedocs.io/en/latest/init-vs-valid-time.html

    flavour = {
        "rules": {
            "latitude": {"name": "latitude"},
            "longitude": {"name": "longitude"},
            "step": {"name": "prediction_timedelta"},
            "date": {"name": "time"},
            "level": {"name": "level"},
        },
        "levtype": "pl",
    }

    fs = XarrayFieldList.from_xarray(ds, flavour=flavour)

    assert_field_list(
        fs,
        2430240,
        "2020-01-01T06:00:00",
        "2021-01-10T12:00:00",
    )


@skip_if_offline
@skip_missing_packages("aiohttp")
def test_inca_one_date() -> None:
    """Test loading and validating the inca_one_date dataset."""
    url = "https://object-store.os-api.cci1.ecmwf.int/ml-tests/test-data/example-inca-one-date.zarr"

    ds = xr.open_zarr(url)
    fs = XarrayFieldList.from_xarray(ds)
    vars = ["DD_10M", "SP_10M", "TD_2M", "TOT_PREC", "T_2M"]

    for i, f in enumerate(fs):
        print(f)
        assert f.metadata("valid_datetime") == "2023-01-01T00:00:00"
        assert f.metadata("step") == 0
        assert f.metadata("number") == 0
        assert f.metadata("variable") == vars[i]

    print(fs[0].datetime())


@skip_if_offline
@skip_missing_packages("gcsfs")
def test_noaa_replay() -> None:
    """Test loading and validating the noaa_replay dataset."""
    ds = xr.open_zarr(
        "gs://noaa-ufs-gefsv13replay/ufs-hr1/1.00-degree/03h-freq/zarr/fv3.zarr",
        storage_options={"token": "anon"},
    )

    flavour = {
        "rules": {
            "latitude": {"name": "grid_yt"},
            "longitude": {"name": "grid_xt"},
            "time": {"name": "time"},
            "level": {"name": "pfull"},
        },
        "levtype": "pl",
    }

    fs = XarrayFieldList.from_xarray(ds, flavour=flavour)

    assert_field_list(
        fs,
        36972870,
        "1993-12-31T18:00:00",
        "1999-06-13T03:00:00",
    )


@skip_if_offline
@skip_missing_packages("planetary_computer", "adlfs")
def test_planetary_computer_conus404() -> None:
    """Test loading and validating the planetary_computer_conus404 dataset."""
    url = "https://planetarycomputer.microsoft.com/api/stac/v1/collections/conus404"
    ds = xr.open_zarr(**name_to_zarr_store(url))

    flavour = {
        "rules": {
            "latitude": {"name": "lat"},
            "longitude": {"name": "lon"},
            "x": {"name": "west_east"},
            "y": {"name": "south_north"},
            "time": {"name": "time"},
        },
    }

    fs = XarrayFieldList.from_xarray(ds, flavour=flavour)

    assert_field_list(
        fs,
        74634912,
        "1979-10-01T00:00:00",
        "2022-09-30T23:00:00",
    )


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()
